@@ -1,5 +1,5 @@
 /*
-*  Copyright (c) 2012-2016, Kaminario Technologies LTD
+*  Copyright (c) 2012-2018, Kaminario Technologies LTD
 *  All rights reserved.
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
@@ -29,7 +29,6 @@
  */
 
 #ifndef __sun
-#define _XOPEN_SOURCE 500
 #ifndef _GNU_SOURCE
 #define _GNU_SOURCE 1
 #endif
@@ -38,6 +37,8 @@
 #include <unistd.h>
 #include <fcntl.h>
 #include <stdlib.h>
+#include <stdio.h>
+#include <stdbool.h>
 #include <string.h>
 #include <errno.h>
 #define __STDC_FORMAT_MACROS 1
@@ -53,7 +54,7 @@
 #include "sg_unaligned.h"
 #include "sg_pr2serr.h"
 
-static const char * version_str = "1.14 20160121";
+static const char * version_str = "1.25 20180625";
 
 #define DEF_BLOCK_SIZE 512
 #define DEF_NUM_BLOCKS (1)
@@ -71,7 +72,9 @@ static struct option long_options[] = {
         {"dpo", no_argument, 0, 'd'},
         {"fua", no_argument, 0, 'f'},
         {"fua_nv", no_argument, 0, 'F'},
+        {"fua-nv", no_argument, 0, 'F'},
         {"group", required_argument, 0, 'g'},
+        {"grpnum", required_argument, 0, 'g'},
         {"help", no_argument, 0, 'h'},
         {"in", required_argument, 0, 'i'},
         {"inc", required_argument, 0, 'C'},
@@ -88,23 +91,25 @@ static struct option long_options[] = {
 };
 
 struct caw_flags {
-        int dpo;
-        int fua;
-        int fua_nv;
+        bool dpo;
+        bool fua;
+        bool fua_nv;
         int group;
         int wrprotect;
 };
 
 struct opts_t {
-        const char * ifn;
-        const char * wfn;
-        int wfn_given;
-        uint64_t lba;
+        bool quiet;
+        bool verbose_given;
+        bool version_given;
+        bool wfn_given;
         int numblocks;
-        int quiet;
         int verbose;
         int timeout;
         int xfer_len;
+        uint64_t lba;
+        const char * ifn;
+        const char * wfn;
         const char * device_name;
         struct caw_flags flags;
 };
@@ -114,12 +119,13 @@ static void
 usage()
 {
         pr2serr("Usage: sg_compare_and_write [--dpo] [--fua] [--fua_nv] "
-                "[--group=GN] [--help]\n"
-                "                            --in=IF [--inw=WF] --lba=LBA "
+                "[--grpnum=GN] [--help]\n"
+                "                            --in=IF|--inc=IF [--inw=WF] "
+                "--lba=LBA "
                 "[--num=NUM]\n"
                 "                            [--quiet] [--timeout=TO] "
                 "[--verbose] [--version]\n"
-                "                            [--wrpotect=WP] [--xferlen=LEN] "
+                "                            [--wrprotect=WP] [--xferlen=LEN] "
                 "DEVICE\n"
                 "  where:\n"
                 "    --dpo|-d            set the dpo bit in cdb (def: "
@@ -128,7 +134,7 @@ usage()
                 "clear)\n"
                 "    --fua_nv|-F         set the fua_nv bit in cdb (def: "
                 "clear)\n"
-                "    --group=GN|-g GN    GN is GROUP NUMBER to set in "
+                "    --grpnum=GN|-g GN    GN is GROUP NUMBER to set in "
                 "cdb (def: 0)\n"
                 "    --help|-h           print out usage message\n"
                 "    --in=IF|-i IF       IF is a file containing a compare "
@@ -136,6 +142,7 @@ usage()
                 "                        optionally a write buffer (when "
                 "--inw=WF is\n"
                 "                        not given)\n"
+                "    --inc=IF|-C IF      The same as the --in option\n"
                 "    --inw=WF|-D WF      WF is a file containing a write "
                 "buffer\n"
                 "    --lba=LBA|-l LBA    LBA of the first block to compare "
@@ -157,15 +164,19 @@ usage()
                 "                            (2 * NUM * 512) or 1024 when "
                 "NUM is 1\n"
                 "\n"
-                "Performs a SCSI COMPARE AND WRITE operation.\n");
+                "Performs a SCSI COMPARE AND WRITE operation. Sends a double "
+                "size\nbuffer, the first half is used to compare what is at "
+                "LBA for NUM\nblocks. If and only if the comparison is "
+                "equal, then the second\nhalf of the buffer is written to "
+                "LBA for NUM blocks.\n");
 }
 
 static int
 parse_args(int argc, char* argv[], struct opts_t * op)
 {
+        bool lba_given = false;
+        bool if_given = false;
         int c;
-        int lba_given = 0;
-        int if_given = 0;
         int64_t ll;
 
         op->numblocks = DEF_NUM_BLOCKS;
@@ -185,27 +196,27 @@ parse_args(int argc, char* argv[], struct opts_t * op)
                 case 'C':
                 case 'i':
                         op->ifn = optarg;
-                        if_given = 1;
+                        if_given = true;
                         break;
                 case 'd':
-                        op->flags.dpo = 1;
+                        op->flags.dpo = true;
                         break;
                 case 'D':
                         op->wfn = optarg;
-                        op->wfn_given = 1;
+                        op->wfn_given = true;
                         break;
                 case 'F':
-                        op->flags.fua_nv = 1;
+                        op->flags.fua_nv = true;
                         break;
                 case 'f':
-                        op->flags.fua = 1;
+                        op->flags.fua = true;
                         break;
                 case 'g':
                         op->flags.group = sg_get_num(optarg);
                         if ((op->flags.group < 0) ||
-                            (op->flags.group > 31))  {
-                                pr2serr("argument to '--group' expected to "
-                                        "be 0 to 31\n");
+                            (op->flags.group > 63))  {
+                                pr2serr("argument to '--grpnum=' expected to "
+                                        "be 0 to 63\n");
                                 goto out_err_no_usage;
                         }
                         break;
@@ -220,7 +231,7 @@ parse_args(int argc, char* argv[], struct opts_t * op)
                                 goto out_err_no_usage;
                         }
                         op->lba = (uint64_t)ll;
-                        lba_given = 1;
+                        lba_given = true;
                         break;
                 case 'n':
                         op->numblocks = sg_get_num(optarg);
@@ -231,7 +242,7 @@ parse_args(int argc, char* argv[], struct opts_t * op)
                         }
                         break;
                 case 'q':
-                        ++op->quiet;
+                        op->quiet = true;
                         break;
                 case 't':
                         op->timeout = sg_get_num(optarg);
@@ -241,11 +252,12 @@ parse_args(int argc, char* argv[], struct opts_t * op)
                         }
                         break;
                 case 'v':
+                        op->verbose_given = true;
                         ++op->verbose;
                         break;
                 case 'V':
-                        pr2serr(ME "version: %s\n", version_str);
-                        exit(0);
+                        op->version_given = true;
+                        break;
                 case 'w':
                         op->flags.wrprotect = sg_get_num(optarg);
                         if (op->flags.wrprotect >> 3) {
@@ -282,11 +294,11 @@ parse_args(int argc, char* argv[], struct opts_t * op)
                 pr2serr("missing device name!\n");
                 goto out_err;
         }
-        if (!if_given) {
+        if (! if_given) {
                 pr2serr("missing input file\n");
                 goto out_err;
         }
-        if (!lba_given) {
+        if (! lba_given) {
                 pr2serr("missing lba\n");
                 goto out_err;
         }
@@ -308,7 +320,7 @@ out_err_no_usage:
 #define WRPROTECT_SHIFT (5)
 
 static int
-sg_build_scsi_cdb(unsigned char * cdbp, unsigned int blocks,
+sg_build_scsi_cdb(uint8_t * cdbp, unsigned int blocks,
                   int64_t start_block, struct caw_flags flags)
 {
         memset(cdbp, 0, COMPARE_AND_WRITE_CDB_SIZE);
@@ -322,23 +334,24 @@ sg_build_scsi_cdb(unsigned char * cdbp, unsigned int blocks,
                 cdbp[1] |= FLAG_FUA_NV;
         sg_put_unaligned_be64((uint64_t)start_block, cdbp + 2);
         /* cdbp[10-12] are reserved */
-        cdbp[13] = (unsigned char)(blocks & 0xff);
-        cdbp[14] = (unsigned char)(flags.group & 0x1f);
+        cdbp[13] = (uint8_t)(blocks & 0xff);
+        cdbp[14] = (uint8_t)(flags.group & 0x1f);
         return 0;
 }
 
 /* Returns 0 for success, SG_LIB_CAT_MISCOMPARE if compare fails,
  * various other SG_LIB_CAT_*, otherwise -1 . */
 static int
-sg_compare_and_write(int sg_fd, unsigned char * buff, int blocks,
-                     int64_t lba, int xfer_len, struct caw_flags flags,
-                     int noisy, int verbose)
+sg_ll_compare_and_write(int sg_fd, uint8_t * buff, int blocks,
+                        int64_t lba, int xfer_len, struct caw_flags flags,
+                        bool noisy, int verbose)
 {
-        int k, sense_cat, valid, slen, res, ret;
-        unsigned char cawCmd[COMPARE_AND_WRITE_CDB_SIZE];
-        unsigned char sense_b[SENSE_BUFF_LEN];
-        struct sg_pt_base * ptvp;
+        bool valid;
+        int k, sense_cat, slen, res, ret;
         uint64_t ull = 0;
+        struct sg_pt_base * ptvp;
+        uint8_t cawCmd[COMPARE_AND_WRITE_CDB_SIZE];
+        uint8_t sense_b[SENSE_BUFF_LEN];
 
         if (sg_build_scsi_cdb(cawCmd, blocks, lba, flags)) {
                 pr2serr(ME "bad cdb build, lba=0x%" PRIx64 ", blocks=%d\n",
@@ -362,14 +375,14 @@ sg_compare_and_write(int sg_fd, unsigned char * buff, int blocks,
         }
         if ((verbose > 2) && (xfer_len > 0)) {
                 pr2serr("    Data-out buffer contents:\n");
-                dStrHexErr((const char *)buff, xfer_len, 1);
+                hex2stderr(buff, xfer_len, 1);
         }
         res = do_scsi_pt(ptvp, sg_fd, DEF_TIMEOUT_SECS, verbose);
-        ret = sg_cmds_process_resp(ptvp, "COMPARE AND WRITE", res, 0,
-                                   sense_b, noisy, verbose,
+        ret = sg_cmds_process_resp(ptvp, "COMPARE AND WRITE", res,
+                                   SG_NO_DATA_IN, sense_b, noisy, verbose,
                                    &sense_cat);
         if (-1 == ret)
-                ;
+                ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
         else if (-2 == ret) {
                 switch (sense_cat) {
                 case SG_LIB_CAT_RECOVERED:
@@ -412,7 +425,7 @@ sg_compare_and_write(int sg_fd, unsigned char * buff, int blocks,
 }
 
 static int
-open_if(const char * fn, int got_stdin)
+open_if(const char * fn, bool got_stdin)
 {
         int fd;
 
@@ -436,13 +449,11 @@ open_if(const char * fn, int got_stdin)
 static int
 open_dev(const char * outf, int verbose)
 {
-        int sg_fd = sg_cmds_open_device(outf, 0 /* rw */, verbose);
-        if (sg_fd < 0) {
+        int sg_fd = sg_cmds_open_device(outf, false /* rw */, verbose);
+
+        if ((sg_fd < 0) && verbose)
                 pr2serr(ME "open error: %s: %s\n", outf,
                         safe_strerror(-sg_fd));
-                return -SG_LIB_FILE_ERROR;
-        }
-
         return sg_fd;
 }
 
@@ -450,13 +461,15 @@ open_dev(const char * outf, int verbose)
 int
 main(int argc, char * argv[])
 {
-        int res, half_xlen, ifn_stdin;
+        bool ifn_stdin;
+        int res, half_xlen, vb;
         int infd = -1;
         int wfd = -1;
         int devfd = -1;
-        unsigned char * wrkBuff = NULL;
-        struct opts_t opts;
+        uint8_t * wrkBuff = NULL;
+        uint8_t * free_wrkBuff = NULL;
         struct opts_t * op;
+        struct opts_t opts;
 
         op = &opts;
         memset(op, 0, sizeof(opts));
@@ -466,7 +479,31 @@ main(int argc, char * argv[])
                 goto out;
         }
 
-        if (op->verbose) {
+#ifdef DEBUG
+        pr2serr("In DEBUG mode, ");
+        if (op->verbose_given && op->version_given) {
+                pr2serr("but override: '-vV' given, zero verbose and "
+                        "continue\n");
+                op->verbose_given = false;
+                op->version_given = false;
+                op->verbose = 0;
+        } else if (! op->verbose_given) {
+                pr2serr("set '-vv'\n");
+                op->verbose = 2;
+        } else
+                pr2serr("keep verbose=%d\n", op->verbose);
+#else
+        if (op->verbose_given && op->version_given)
+                pr2serr("Not in DEBUG mode, so '-vV' has no special "
+                        "action\n");
+#endif
+        if (op->version_given) {
+                pr2serr(ME "version: %s\n", version_str);
+                return 0;
+        }
+        vb = op->verbose;
+
+        if (vb) {
                 pr2serr("Running COMPARE AND WRITE command with the "
                         "following options:\n  in=%s ", op->ifn);
                 if (op->wfn_given)
@@ -487,23 +524,24 @@ main(int argc, char * argv[])
                         res = SG_LIB_FILE_ERROR;
                         goto out;
                 }
-                wfd = open_if(op->wfn, 0);
+                wfd = open_if(op->wfn, false);
                 if (wfd < 0) {
                         res = -wfd;
                         goto out;
                 }
         }
 
-        devfd = open_dev(op->device_name, op->verbose);
+        devfd = open_dev(op->device_name, vb);
         if (devfd < 0) {
-                res = -devfd;
+                res = sg_convert_errno(-devfd);
                 goto out;
         }
 
-        wrkBuff = (unsigned char *)malloc(op->xfer_len);
-        if (0 == wrkBuff) {
+        wrkBuff = (uint8_t *)sg_memalign(op->xfer_len, 0, &free_wrkBuff,
+                                         vb > 3);
+        if (NULL == wrkBuff) {
                 pr2serr("Not enough user memory\n");
-                res = SG_LIB_CAT_OTHER;
+                res = sg_convert_errno(ENOMEM);
                 goto out;
         }
 
@@ -538,10 +576,9 @@ main(int argc, char * argv[])
                         goto out;
                 }
         }
-        res = sg_compare_and_write(devfd, wrkBuff, op->numblocks, op->lba,
-                op->xfer_len, op->flags, !op->quiet, op->verbose);
-
-out:
+        res = sg_ll_compare_and_write(devfd, wrkBuff, op->numblocks, op->lba,
+                                      op->xfer_len, op->flags, ! op->quiet,
+                                      vb);
         if (0 != res) {
                 char b[80];
 
@@ -551,20 +588,24 @@ out:
                 case SG_LIB_FILE_ERROR:
                         break;  /* already reported */
                 default:
-                        sg_get_category_sense_str(res, sizeof(b), b,
-                                                  op->verbose);
+                        sg_get_category_sense_str(res, sizeof(b), b, vb);
                         pr2serr(ME "SCSI COMPARE AND WRITE: %s\n", b);
                         break;
                 }
         }
-
-        if (wrkBuff)
-                free(wrkBuff);
+out:
+        if (free_wrkBuff)
+                free(free_wrkBuff);
         if ((infd >= 0) && (! ifn_stdin))
                 close(infd);
         if (wfd >= 0)
                 close(wfd);
         if (devfd >= 0)
                 close(devfd);
+        if (0 == op->verbose) {
+                if (! sg_if_can2stderr("sg_compare_and_write failed: ", res))
+                        pr2serr("Some error occurred, try again with '-v' "
+                                "or '-vv' for more information\n");
+        }
         return res;
 }
