@@ -1,5 +1,5 @@
 /* A utility program originally written for the Linux OS SCSI subsystem.
- *  Copyright (C) 1999-2013 D. Gilbert
+ *  Copyright (C) 1999-2018 D. Gilbert
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
  *  the Free Software Foundation; either version 2, or (at your option)
@@ -11,15 +11,17 @@
  */
 
 
-#define _XOPEN_SOURCE 500
+#define _XOPEN_SOURCE 600
 #ifndef _GNU_SOURCE
-#define _GNU_SOURCE
+#define _GNU_SOURCE 1
 #endif
 
 #include <unistd.h>
 #include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
+#include <stdbool.h>
 #include <string.h>
 #include <errno.h>
 #include <getopt.h>
@@ -36,84 +38,93 @@
 #endif
 #include "sg_lib.h"
 #include "sg_io_linux.h"
+#include "sg_unaligned.h"
+#include "sg_pr2serr.h"
 
 #define RB_MODE_DESC 3
 #define RB_MODE_DATA 2
+#define RB_MODE_ECHO_DESC 0xb
+#define RB_MODE_ECHO_DATA 0xa
 #define RB_DESC_LEN 4
 #define RB_DEF_SIZE (200*1024*1024)
 #define RB_OPCODE 0x3C
 #define RB_CMD_LEN 10
 
-/* #define SG_DEBUG */
-
 #ifndef SG_FLAG_MMAP_IO
 #define SG_FLAG_MMAP_IO 4
 #endif
 
 
-static const char * version_str = "4.89 20130507";
+static const char * version_str = "5.05 20180627";
 
 static struct option long_options[] = {
-        {"buffer", 1, 0, 'b'},
-        {"dio", 0, 0, 'd'},
-        {"help", 0, 0, 'h'},
-        {"mmap", 0, 0, 'm'},
-        {"new", 0, 0, 'N'},
-        {"old", 0, 0, 'O'},
-        {"quick", 0, 0, 'q'},
-        {"size", 1, 0, 's'},
-        {"time", 0, 0, 't'},
-        {"verbose", 0, 0, 'v'},
-        {"version", 0, 0, 'V'},
+        {"buffer", required_argument, 0, 'b'},
+        {"dio", no_argument, 0, 'd'},
+        {"echo", no_argument, 0, 'e'},
+        {"help", no_argument, 0, 'h'},
+        {"mmap", no_argument, 0, 'm'},
+        {"new", no_argument, 0, 'N'},
+        {"old", no_argument, 0, 'O'},
+        {"quick", no_argument, 0, 'q'},
+        {"size", required_argument, 0, 's'},
+        {"time", no_argument, 0, 't'},
+        {"verbose", no_argument, 0, 'v'},
+        {"version", no_argument, 0, 'V'},
         {0, 0, 0, 0},
 };
 
 struct opts_t {
+    bool do_dio;
+    bool do_echo;
+    bool do_mmap;
+    bool do_quick;
+    bool do_time;
+    bool verbose_given;
+    bool version_given;
+    bool opt_new;
     int do_buffer;
-    int do_dio;
     int do_help;
-    int do_mmap;
-    int do_quick;
+    int verbose;
     int64_t do_size;
-    int do_time;
-    int do_verbose;
-    int do_version;
     const char * device_name;
-    int opt_new;
 };
 
 
 static void
 usage()
 {
-    fprintf(stderr, "Usage: sg_rbuf [--buffer=EACH] [--dio] [--help] "
-            "[--mmap] [--quick]\n"
-            "               [--size=OVERALL] [--time] [--verbose] "
-            "[--version] DEVICE\n");
-    fprintf(stderr, "  where:\n"
+    pr2serr("Usage: sg_rbuf [--buffer=EACH] [--dio] [--echo] "
+            "[--help] [--mmap]\n"
+            "               [--quick] [--size=OVERALL] [--time] [--verbose] "
+            "[--version]\n"
+            "               SG_DEVICE\n");
+    pr2serr("  where:\n"
             "    --buffer=EACH|-b EACH    buffer size to use (in bytes)\n"
             "    --dio|-d        requests dio ('-q' overrides it)\n"
+            "    --echo|-e       use echo buffer (def: use data mode)\n"
             "    --help|-h       print usage message then exit\n"
             "    --mmap|-m       requests mmap-ed IO (overrides -q, -d)\n"
             "    --quick|-q      quick, don't xfer to user space\n");
-    fprintf(stderr,
-            "    --size=OVERALL|-s OVERALL    total size to read (in bytes)\n"
+    pr2serr("    --size=OVERALL|-s OVERALL    total size to read (in bytes)\n"
             "                    default: 200 MiB\n"
             "    --time|-t       time the data transfer\n"
             "    --verbose|-v    increase verbosity (more debug)\n"
+            "    --old|-O        use old interface (use as first option)\n"
             "    --version|-V    print version string then exit\n\n"
-            "Use SCSI READ BUFFER command (data mode, buffer id 0) "
-            "repeatedly\n");
+            "Use SCSI READ BUFFER command (data or echo buffer mode, buffer "
+            "id 0)\nrepeatedly. This utility only works with Linux sg "
+            "devices.\n");
 }
 
 static void
 usage_old()
 {
     printf("Usage: sg_rbuf [-b=EACH_KIB] [-d] [-m] [-q] [-s=OVERALL_MIB] "
-           "[-t] [-v] [-V]\n               DEVICE\n");
+           "[-t] [-v] [-V]\n               SG_DEVICE\n");
     printf("  where:\n");
     printf("    -b=EACH_KIB    num is buffer size to use (in KiB)\n");
     printf("    -d       requests dio ('-q' overrides it)\n");
+    printf("    -e       use echo buffer (def: use data mode)\n");
     printf("    -m       requests mmap-ed IO (overrides -q, -d)\n");
     printf("    -q       quick, don't xfer to user space\n");
     printf("    -s=OVERALL_MIB    num is total size to read (in MiB) "
@@ -121,22 +132,24 @@ usage_old()
     printf("             maximum total size is 4000 MiB\n");
     printf("    -t       time the data transfer\n");
     printf("    -v       increase verbosity (more debug)\n");
+    printf("    -N|--new use new interface\n");
     printf("    -V       print version string then exit\n\n");
-    printf("Use SCSI READ BUFFER command (data mode, buffer id 0) "
-           "repeatedly\n");
+    printf("Use SCSI READ BUFFER command (data or echo buffer mode, buffer "
+           "id 0)\nrepeatedly. This utility only works with Linux sg "
+            "devices.\n");
 }
 
 static void
-usage_for(const struct opts_t * optsp)
+usage_for(const struct opts_t * op)
 {
-    if (optsp->opt_new)
+    if (op->opt_new)
         usage();
     else
         usage_old();
 }
 
 static int
-process_cl_new(struct opts_t * optsp, int argc, char * argv[])
+new_parse_cmd_line(struct opts_t * op, int argc, char * argv[])
 {
     int c, n;
     int64_t nn;
@@ -144,7 +157,7 @@ process_cl_new(struct opts_t * optsp, int argc, char * argv[])
     while (1) {
         int option_index = 0;
 
-        c = getopt_long(argc, argv, "b:dhmNOqs:tvV", long_options,
+        c = getopt_long(argc, argv, "b:dehmNOqs:tvV", long_options,
                         &option_index);
         if (c == -1)
             break;
@@ -153,66 +166,69 @@ process_cl_new(struct opts_t * optsp, int argc, char * argv[])
         case 'b':
             n = sg_get_num(optarg);
             if (n < 0) {
-                fprintf(stderr, "bad argument to '--buffer'\n");
-                usage_for(optsp);
+                pr2serr("bad argument to '--buffer'\n");
+                usage_for(op);
                 return SG_LIB_SYNTAX_ERROR;
             }
-            optsp->do_buffer = n;
+            op->do_buffer = n;
             break;
         case 'd':
-            ++optsp->do_dio;
+            op->do_dio = true;
+            break;
+        case 'e':
+            op->do_echo = true;
             break;
         case 'h':
         case '?':
-            ++optsp->do_help;
+            ++op->do_help;
             break;
         case 'm':
-            ++optsp->do_mmap;
+            op->do_mmap = true;
             break;
         case 'N':
             break;      /* ignore */
         case 'O':
-            optsp->opt_new = 0;
+            op->opt_new = false;
             return 0;
         case 'q':
-            ++optsp->do_quick;
+            op->do_quick = true;
             break;
         case 's':
            nn = sg_get_llnum(optarg);
            if (nn < 0) {
-                fprintf(stderr, "bad argument to '--size'\n");
-                usage_for(optsp);
+                pr2serr("bad argument to '--size'\n");
+                usage_for(op);
                 return SG_LIB_SYNTAX_ERROR;
             }
-            optsp->do_size = nn;
+            op->do_size = nn;
             break;
         case 't':
-            ++optsp->do_time;
+            op->do_time = true;
             break;
         case 'v':
-            ++optsp->do_verbose;
+            op->verbose_given = true;
+            ++op->verbose;
             break;
         case 'V':
-            ++optsp->do_version;
+            op->version_given = true;
             break;
         default:
-            fprintf(stderr, "unrecognised option code %c [0x%x]\n", c, c);
-            if (optsp->do_help)
+            pr2serr("unrecognised option code %c [0x%x]\n", c, c);
+            if (op->do_help)
                 break;
-            usage_for(optsp);
+            usage_for(op);
             return SG_LIB_SYNTAX_ERROR;
         }
     }
     if (optind < argc) {
-        if (NULL == optsp->device_name) {
-            optsp->device_name = argv[optind];
+        if (NULL == op->device_name) {
+            op->device_name = argv[optind];
             ++optind;
         }
         if (optind < argc) {
             for (; optind < argc; ++optind)
-                fprintf(stderr, "Unexpected extra argument: %s\n",
-                        argv[optind]);
-            usage_for(optsp);
+                pr2serr("Unexpected extra argument: %s\n", argv[optind]);
+            usage_for(op);
             return SG_LIB_SYNTAX_ERROR;
         }
     }
@@ -220,9 +236,10 @@ process_cl_new(struct opts_t * optsp, int argc, char * argv[])
 }
 
 static int
-process_cl_old(struct opts_t * optsp, int argc, char * argv[])
+old_parse_cmd_line(struct opts_t * op, int argc, char * argv[])
 {
-    int k, jmp_out, plen, num;
+    bool jmp_out;
+    int k, plen, num;
     int64_t nn;
     const char * cp;
 
@@ -232,37 +249,41 @@ process_cl_old(struct opts_t * optsp, int argc, char * argv[])
         if (plen <= 0)
             continue;
         if ('-' == *cp) {
-            for (--plen, ++cp, jmp_out = 0; plen > 0; --plen, ++cp) {
+            for (--plen, ++cp, jmp_out = false; plen > 0; --plen, ++cp) {
                 switch (*cp) {
                 case 'd':
-                    ++optsp->do_dio;
+                    op->do_dio = true;
+                    break;
+                case 'e':
+                    op->do_echo = true;
                     break;
                 case 'h':
                 case '?':
-                    ++optsp->do_help;
+                    ++op->do_help;
                     break;
                 case 'm':
-                    ++optsp->do_mmap;
+                    op->do_mmap = true;
                     break;
                 case 'N':
-                    optsp->opt_new = 1;
+                    op->opt_new = true;
                     return 0;
                 case 'O':
                     break;
                 case 'q':
-                    ++optsp->do_quick;
+                    op->do_quick = true;
                     break;
                 case 't':
-                    ++optsp->do_time;
+                    op->do_time = true;
                     break;
                 case 'v':
-                    ++optsp->do_verbose;
+                    op->verbose_given = true;
+                    ++op->verbose;
                     break;
                 case 'V':
-                    ++optsp->do_version;
+                    op->version_given = true;
                     break;
                 default:
-                    jmp_out = 1;
+                    jmp_out = true;
                     break;
                 }
                 if (jmp_out)
@@ -271,36 +292,36 @@ process_cl_old(struct opts_t * optsp, int argc, char * argv[])
             if (plen <= 0)
                 continue;
             if (0 == strncmp("b=", cp, 2)) {
-                num = sscanf(cp + 2, "%d", &optsp->do_buffer);
-                if ((1 != num) || (optsp->do_buffer <= 0)) {
+                num = sscanf(cp + 2, "%d", &op->do_buffer);
+                if ((1 != num) || (op->do_buffer <= 0)) {
                     printf("Couldn't decode number after 'b=' option\n");
-                    usage_for(optsp);
+                    usage_for(op);
                     return SG_LIB_SYNTAX_ERROR;
                 }
-                optsp->do_buffer *= 1024;
+                op->do_buffer *= 1024;
             }
             else if (0 == strncmp("s=", cp, 2)) {
                 nn = sg_get_llnum(optarg);
                 if (nn < 0) {
                     printf("Couldn't decode number after 's=' option\n");
-                    usage_for(optsp);
+                    usage_for(op);
                     return SG_LIB_SYNTAX_ERROR;
                 }
-                optsp->do_size = nn;
-                optsp->do_size *= 1024 * 1024;
+                op->do_size = nn;
+                op->do_size *= 1024 * 1024;
             } else if (0 == strncmp("-old", cp, 4))
                 ;
             else if (jmp_out) {
-                fprintf(stderr, "Unrecognized option: %s\n", cp);
-                usage_for(optsp);
+                pr2serr("Unrecognized option: %s\n", cp);
+                usage_for(op);
                 return SG_LIB_SYNTAX_ERROR;
             }
-        } else if (0 == optsp->device_name)
-            optsp->device_name = cp;
+        } else if (0 == op->device_name)
+            op->device_name = cp;
         else {
-            fprintf(stderr, "too many arguments, got: %s, not expecting: "
-                    "%s\n", optsp->device_name, cp);
-            usage_for(optsp);
+            pr2serr("too many arguments, got: %s, not expecting: %s\n",
+                    op->device_name, cp);
+            usage_for(op);
             return SG_LIB_SYNTAX_ERROR;
         }
     }
@@ -308,22 +329,22 @@ process_cl_old(struct opts_t * optsp, int argc, char * argv[])
 }
 
 static int
-process_cl(struct opts_t * optsp, int argc, char * argv[])
+parse_cmd_line(struct opts_t * op, int argc, char * argv[])
 {
     int res;
     char * cp;
 
     cp = getenv("SG3_UTILS_OLD_OPTS");
     if (cp) {
-        optsp->opt_new = 0;
-        res = process_cl_old(optsp, argc, argv);
-        if ((0 == res) && optsp->opt_new)
-            res = process_cl_new(optsp, argc, argv);
+        op->opt_new = false;
+        res = old_parse_cmd_line(op, argc, argv);
+        if ((0 == res) && op->opt_new)
+            res = new_parse_cmd_line(op, argc, argv);
     } else {
-        optsp->opt_new = 1;
-        res = process_cl_new(optsp, argc, argv);
-        if ((0 == res) && (0 == optsp->opt_new))
-            res = process_cl_old(optsp, argc, argv);
+        op->opt_new = true;
+        res = new_parse_cmd_line(op, argc, argv);
+        if ((0 == res) && (! op->opt_new))
+            res = old_parse_cmd_line(op, argc, argv);
     }
     return res;
 }
@@ -332,113 +353,149 @@ process_cl(struct opts_t * optsp, int argc, char * argv[])
 int
 main(int argc, char * argv[])
 {
-    int sg_fd, res, j;
-    unsigned int k, num;
-    unsigned char rbCmdBlk [RB_CMD_LEN];
-    unsigned char * rbBuff = NULL;
-    void * rawp = NULL;
-    unsigned char sense_buffer[32];
+#ifdef DEBUG
+    bool clear = true;
+#endif
+    bool dio_incomplete = false;
+    int sg_fd, res, j, err;
     int buf_capacity = 0;
     int buf_size = 0;
+    size_t psz;
+    unsigned int k, num;
     int64_t total_size = RB_DEF_SIZE;
-    size_t psz = getpagesize();
-    int dio_incomplete = 0;
+    struct opts_t * op;
+    uint8_t * rbBuff = NULL;
+    void * rawp = NULL;
+    uint8_t sense_buffer[32];
+    uint8_t rb_cdb [RB_CMD_LEN];
     struct sg_io_hdr io_hdr;
     struct timeval start_tm, end_tm;
-#ifdef SG_DEBUG
-    int clear = 1;
-#endif
     struct opts_t opts;
 
-    memset(&opts, 0, sizeof(opts));
-    res = process_cl(&opts, argc, argv);
+#if defined(HAVE_SYSCONF) && defined(_SC_PAGESIZE)
+    psz = sysconf(_SC_PAGESIZE); /* POSIX.1 (was getpagesize()) */
+#else
+    psz = 4096;     /* give up, pick likely figure */
+#endif
+    op = &opts;
+    memset(op, 0, sizeof(opts));
+    res = parse_cmd_line(op, argc, argv);
     if (res)
         return SG_LIB_SYNTAX_ERROR;
-    if (opts.do_help) {
-        usage_for(&opts);
+    if (op->do_help) {
+        usage_for(op);
         return 0;
     }
-    if (opts.do_version) {
-        fprintf(stderr, "Version string: %s\n", version_str);
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (op->verbose_given && op->version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        op->verbose_given = false;
+        op->version_given = false;
+        op->verbose = 0;
+    } else if (! op->verbose_given) {
+        pr2serr("set '-vv'\n");
+        op->verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", op->verbose);
+#else
+    if (op->verbose_given && op->version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (op->version_given) {
+        pr2serr("Version string: %s\n", version_str);
         return 0;
     }
 
-    if (NULL == opts.device_name) {
-        fprintf(stderr, "No DEVICE argument given\n");
-        usage_for(&opts);
+    if (NULL == op->device_name) {
+        pr2serr("No DEVICE argument given\n\n");
+        usage_for(op);
         return SG_LIB_SYNTAX_ERROR;
     }
 
-    if (opts.do_buffer > 0)
-        buf_size = opts.do_buffer;
-    if (opts.do_size > 0)
-        total_size = opts.do_size;
+    if (op->do_buffer > 0)
+        buf_size = op->do_buffer;
+    if (op->do_size > 0)
+        total_size = op->do_size;
 
-    sg_fd = open(opts.device_name, O_RDONLY | O_NONBLOCK);
+    sg_fd = open(op->device_name, O_RDONLY | O_NONBLOCK);
     if (sg_fd < 0) {
+        err = errno;
         perror("device open error");
-        return SG_LIB_FILE_ERROR;
+        return sg_convert_errno(err);
     }
-    /* Don't worry, being very careful not to write to a none-sg file ... */
-    if (opts.do_mmap) {
-        opts.do_dio = 0;
-        opts.do_quick = 0;
+    if (op->do_mmap) {
+        op->do_dio = false;
+        op->do_quick = false;
     }
     if (NULL == (rawp = malloc(512))) {
         printf("out of memory (query)\n");
         return SG_LIB_CAT_OTHER;
     }
-    rbBuff = (unsigned char *)rawp;
+    rbBuff = (uint8_t *)rawp;
 
-    memset(rbCmdBlk, 0, RB_CMD_LEN);
-    rbCmdBlk[0] = RB_OPCODE;
-    rbCmdBlk[1] = RB_MODE_DESC; /* data mode, buffer id 0 */
-    rbCmdBlk[8] = RB_DESC_LEN;
+    memset(rb_cdb, 0, RB_CMD_LEN);
+    rb_cdb[0] = RB_OPCODE;
+    rb_cdb[1] = op->do_echo ? RB_MODE_ECHO_DESC : RB_MODE_DESC;
+    rb_cdb[8] = RB_DESC_LEN;
     memset(&io_hdr, 0, sizeof(struct sg_io_hdr));
     io_hdr.interface_id = 'S';
-    io_hdr.cmd_len = sizeof(rbCmdBlk);
+    io_hdr.cmd_len = sizeof(rb_cdb);
     io_hdr.mx_sb_len = sizeof(sense_buffer);
     io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
     io_hdr.dxfer_len = RB_DESC_LEN;
     io_hdr.dxferp = rbBuff;
-    io_hdr.cmdp = rbCmdBlk;
+    io_hdr.cmdp = rb_cdb;
     io_hdr.sbp = sense_buffer;
     io_hdr.timeout = 60000;     /* 60000 millisecs == 60 seconds */
-    if (opts.do_verbose) {
-        fprintf(stderr, "    Read buffer cdb: ");
+    if (op->verbose) {
+        pr2serr("    Read buffer (%sdescriptor) cdb: ",
+                (op->do_echo ? "echo " : ""));
         for (k = 0; k < RB_CMD_LEN; ++k)
-            fprintf(stderr, "%02x ", rbCmdBlk[k]);
-        fprintf(stderr, "\n");
+            pr2serr("%02x ", rb_cdb[k]);
+        pr2serr("\n");
     }
 
     /* do normal IO to find RB size (not dio or mmap-ed at this stage) */
     if (ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
         perror("SG_IO READ BUFFER descriptor error");
-        if (rawp) free(rawp);
+        if (rawp)
+            free(rawp);
         return SG_LIB_CAT_OTHER;
     }
 
-    if (opts.do_verbose > 2)
-        fprintf(stderr, "      duration=%u ms\n", io_hdr.duration);
+    if (op->verbose > 2)
+        pr2serr("      duration=%u ms\n", io_hdr.duration);
     /* now for the error processing */
     res = sg_err_category3(&io_hdr);
     switch (res) {
     case SG_LIB_CAT_RECOVERED:
         sg_chk_n_print3("READ BUFFER descriptor, continuing", &io_hdr,
-                        opts.do_verbose > 1);
-        /* fall through */
+                        op->verbose > 1);
+#if defined(__GNUC__)
+#if (__GNUC__ >= 7)
+        __attribute__((fallthrough));
+        /* FALL THROUGH */
+#endif
+#endif
     case SG_LIB_CAT_CLEAN:
         break;
     default: /* won't bother decoding other categories */
         sg_chk_n_print3("READ BUFFER descriptor error", &io_hdr,
-                        opts.do_verbose > 1);
+                        op->verbose > 1);
         if (rawp) free(rawp);
         return (res >= 0) ? res : SG_LIB_CAT_OTHER;
     }
 
-    buf_capacity = ((rbBuff[1] << 16) | (rbBuff[2] << 8) | rbBuff[3]);
-    printf("READ BUFFER reports: buffer capacity=%d, offset boundary=%d\n",
-           buf_capacity, (int)rbBuff[0]);
+    if (op->do_echo) {
+        buf_capacity = 0x1fff & sg_get_unaligned_be16(rbBuff + 2);
+        printf("READ BUFFER reports: echo buffer capacity=%d\n",
+               buf_capacity);
+    } else {
+        buf_capacity = sg_get_unaligned_be24(rbBuff + 1);
+        printf("READ BUFFER reports: buffer capacity=%d, offset "
+               "boundary=%d\n", buf_capacity, (int)rbBuff[0]);
+    }
 
     if (0 == buf_size)
         buf_size = buf_capacity;
@@ -453,108 +510,105 @@ main(int argc, char * argv[])
         rawp = NULL;
     }
 
-    if (! opts.do_dio) {
+    if (! op->do_dio) {
         k = buf_size;
-        if (opts.do_mmap && (0 != (k % psz)))
+        if (op->do_mmap && (0 != (k % psz)))
             k = ((k / psz) + 1) * psz;  /* round up to page size */
         res = ioctl(sg_fd, SG_SET_RESERVED_SIZE, &k);
         if (res < 0)
             perror("SG_SET_RESERVED_SIZE error");
     }
 
-    if (opts.do_mmap) {
-        rbBuff = (unsigned char *)mmap(NULL, buf_size, PROT_READ, MAP_SHARED,
+    if (op->do_mmap) {
+        rbBuff = (uint8_t *)mmap(NULL, buf_size, PROT_READ, MAP_SHARED,
                                        sg_fd, 0);
         if (MAP_FAILED == rbBuff) {
             if (ENOMEM == errno) {
-                fprintf(stderr, "mmap() out of memory, try a smaller "
-                       "buffer size than %d bytes\n", buf_size);
-                if (opts.opt_new)
-                    fprintf(stderr, "    [with '--buffer=EACH' where EACH "
-                            "is in bytes]\n");
+                pr2serr("mmap() out of memory, try a smaller buffer size "
+                        "than %d bytes\n", buf_size);
+                if (op->opt_new)
+                    pr2serr("    [with '--buffer=EACH' where EACH is in "
+                            "bytes]\n");
                 else
-                    fprintf(stderr, "    [with '-b=EACH' where EACH is in "
-                            "KiB]\n");
+                    pr2serr("    [with '-b=EACH' where EACH is in KiB]\n");
             } else
                 perror("error using mmap()");
             return SG_LIB_CAT_OTHER;
         }
     }
     else { /* non mmap-ed IO */
-        rawp = (unsigned char *)malloc(buf_size + (opts.do_dio ? psz : 0));
+        rawp = (uint8_t *)malloc(buf_size + (op->do_dio ? psz : 0));
         if (NULL == rawp) {
             printf("out of memory (data)\n");
             return SG_LIB_CAT_OTHER;
         }
-        if (opts.do_dio)    /* align to page boundary */
-            rbBuff= (unsigned char *)(((unsigned long)rawp + psz - 1) &
+        /* perhaps use posix_memalign() instead */
+        if (op->do_dio)    /* align to page boundary */
+            rbBuff= (uint8_t *)(((sg_uintptr_t)rawp + psz - 1) &
                                       (~(psz - 1)));
         else
-            rbBuff = (unsigned char *)rawp;
+            rbBuff = (uint8_t *)rawp;
     }
 
     num = total_size / buf_size;
-    if (opts.do_time) {
+    if (op->do_time) {
         start_tm.tv_sec = 0;
         start_tm.tv_usec = 0;
         gettimeofday(&start_tm, NULL);
     }
     /* main data reading loop */
     for (k = 0; k < num; ++k) {
-        memset(rbCmdBlk, 0, RB_CMD_LEN);
-        rbCmdBlk[0] = RB_OPCODE;
-        rbCmdBlk[1] = RB_MODE_DATA;
-        rbCmdBlk[6] = 0xff & (buf_size >> 16);
-        rbCmdBlk[7] = 0xff & (buf_size >> 8);
-        rbCmdBlk[8] = 0xff & buf_size;
-#ifdef SG_DEBUG
+        memset(rb_cdb, 0, RB_CMD_LEN);
+        rb_cdb[0] = RB_OPCODE;
+        rb_cdb[1] = op->do_echo ? RB_MODE_ECHO_DATA : RB_MODE_DATA;
+        sg_put_unaligned_be24((uint32_t)buf_size, rb_cdb + 6);
+#ifdef DEBUG
         memset(rbBuff, 0, buf_size);
 #endif
 
         memset(&io_hdr, 0, sizeof(struct sg_io_hdr));
         io_hdr.interface_id = 'S';
-        io_hdr.cmd_len = sizeof(rbCmdBlk);
+        io_hdr.cmd_len = sizeof(rb_cdb);
         io_hdr.mx_sb_len = sizeof(sense_buffer);
         io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
         io_hdr.dxfer_len = buf_size;
-        if (! opts.do_mmap)
+        if (! op->do_mmap)
             io_hdr.dxferp = rbBuff;
-        io_hdr.cmdp = rbCmdBlk;
+        io_hdr.cmdp = rb_cdb;
         io_hdr.sbp = sense_buffer;
         io_hdr.timeout = 20000;     /* 20000 millisecs == 20 seconds */
         io_hdr.pack_id = k;
-        if (opts.do_mmap)
+        if (op->do_mmap)
             io_hdr.flags |= SG_FLAG_MMAP_IO;
-        else if (opts.do_dio)
+        else if (op->do_dio)
             io_hdr.flags |= SG_FLAG_DIRECT_IO;
-        else if (opts.do_quick)
+        else if (op->do_quick)
             io_hdr.flags |= SG_FLAG_NO_DXFER;
-        if (opts.do_verbose > 1) {
-            fprintf(stderr, "    Read buffer cdb: ");
+        if (op->verbose > 1) {
+            pr2serr("    Read buffer (%sdata) cdb: ",
+                    (op->do_echo ? "echo " : ""));
             for (j = 0; j < RB_CMD_LEN; ++j)
-                fprintf(stderr, "%02x ", rbCmdBlk[j]);
-            fprintf(stderr, "\n");
+                pr2serr("%02x ", rb_cdb[j]);
+            pr2serr("\n");
         }
 
         if (ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
             if (ENOMEM == errno) {
-                fprintf(stderr, "SG_IO data: out of memory, try a smaller "
-                       "buffer size than %d bytes\n", buf_size);
-                if (opts.opt_new)
-                    fprintf(stderr, "    [with '--buffer=EACH' where EACH "
-                            "is in bytes]\n");
+                pr2serr("SG_IO data: out of memory, try a smaller buffer "
+                        "size than %d bytes\n", buf_size);
+                if (op->opt_new)
+                    pr2serr("    [with '--buffer=EACH' where EACH is in "
+                            "bytes]\n");
                 else
-                    fprintf(stderr, "    [with '-b=EACH' where EACH is in "
-                            "KiB]\n");
+                    pr2serr("    [with '-b=EACH' where EACH is in KiB]\n");
             } else
                 perror("SG_IO READ BUFFER data error");
             if (rawp) free(rawp);
             return SG_LIB_CAT_OTHER;
         }
 
-        if (opts.do_verbose > 2)
-            fprintf(stderr, "      duration=%u ms\n",
-                    io_hdr.duration);
+        if (op->verbose > 2)
+            pr2serr("      duration=%u ms\n", io_hdr.duration);
         /* now for the error processing */
         res = sg_err_category3(&io_hdr);
         switch (res) {
@@ -562,30 +616,30 @@ main(int argc, char * argv[])
             break;
         case SG_LIB_CAT_RECOVERED:
             sg_chk_n_print3("READ BUFFER data, continuing", &io_hdr,
-                            opts.do_verbose > 1);
+                            op->verbose > 1);
             break;
         default: /* won't bother decoding other categories */
             sg_chk_n_print3("READ BUFFER data error", &io_hdr,
-                            opts.do_verbose > 1);
+                            op->verbose > 1);
             if (rawp) free(rawp);
             return (res >= 0) ? res : SG_LIB_CAT_OTHER;
         }
-        if (opts.do_dio &&
+        if (op->do_dio &&
             ((io_hdr.info & SG_INFO_DIRECT_IO_MASK) != SG_INFO_DIRECT_IO))
-            dio_incomplete = 1;    /* flag that dio not done (completely) */
+            dio_incomplete = true;  /* flag that dio not done (completely) */
 
-#ifdef SG_DEBUG
+#ifdef DEBUG
         if (clear) {
             for (j = 0; j < buf_size; ++j) {
                 if (rbBuff[j] != 0) {
-                    clear = 0;
+                    clear = false;
                     break;
                 }
             }
         }
 #endif
     }
-    if ((opts.do_time) && (start_tm.tv_sec || start_tm.tv_usec)) {
+    if (op->do_time && (start_tm.tv_sec || start_tm.tv_usec)) {
         struct timeval res_tm;
         double a, b;
 
@@ -601,24 +655,27 @@ main(int argc, char * argv[])
         b = (double)buf_size * num;
         printf("time to read data from buffer was %d.%06d secs",
                (int)res_tm.tv_sec, (int)res_tm.tv_usec);
-        if ((a > 0.00001) && (b > 511))
-            printf(", %.2f MB/sec\n", b / (a * 1000000.0));
-        else
-            printf("\n");
+        if (a > 0.00001) {
+            if (b > 511)
+                printf(", %.2f MB/sec", b / (a * 1000000.0));
+            printf(", %.2f IOPS", num / a);
+        }
+        printf("\n");
     }
     if (dio_incomplete)
         printf(">> direct IO requested but not done\n");
-    printf("Read %"PRId64" MiB (actual: %"PRId64" bytes), buffer size=%d KiB "
-           "(%d bytes)\n", (total_size / (1024 * 1024)),
+    printf("Read %" PRId64 " MiB (actual: %" PRId64 " bytes), buffer "
+           "size=%d KiB (%d bytes)\n", (total_size / (1024 * 1024)),
            (int64_t)num * buf_size, buf_size / 1024, buf_size);
 
     if (rawp) free(rawp);
     res = close(sg_fd);
     if (res < 0) {
+        err = errno;
         perror("close error");
-        return SG_LIB_FILE_ERROR;
+        return sg_convert_errno(err);
     }
-#ifdef SG_DEBUG
+#ifdef DEBUG
     if (clear)
         printf("read buffer always zero\n");
     else
