@@ -1,5 +1,5 @@
 /* A utility program for the Linux OS SCSI generic ("sg") device driver.
-*  Copyright (C) 2001 - 2016 D. Gilbert
+*  Copyright (C) 2001 - 2018 D. Gilbert
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
@@ -18,7 +18,7 @@
 
 */
 
-#define _XOPEN_SOURCE 500
+#define _XOPEN_SOURCE 600
 #ifndef _GNU_SOURCE
 #define _GNU_SOURCE 1
 #endif
@@ -27,6 +27,8 @@
 #include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
+#include <stdbool.h>
 #include <string.h>
 #include <signal.h>
 #include <ctype.h>
@@ -34,9 +36,11 @@
 #define __STDC_FORMAT_MACROS 1
 #include <inttypes.h>
 #include <sys/ioctl.h>
-#include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/sysmacros.h>
+#ifndef major
+#include <sys/types.h>
+#endif
 #include <sys/mman.h>
 #include <sys/time.h>
 #include <linux/major.h>
@@ -44,13 +48,14 @@
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
+
 #include "sg_lib.h"
 #include "sg_io_linux.h"
 #include "sg_unaligned.h"
 #include "sg_pr2serr.h"
 
 
-static const char * version_str = "1.25 20160121";
+static const char * version_str = "1.34 20180811";
 
 #define DEF_BLOCK_SIZE 512
 #define DEF_BLOCKS_PER_TRANSFER 128
@@ -90,7 +95,9 @@ static int verbose = 0;
 
 static const char * proc_allow_dio = "/proc/scsi/sg/allow_dio";
 
-static void install_handler (int sig_num, void (*sig_handler) (int sig))
+
+static void
+install_handler (int sig_num, void (*sig_handler) (int sig))
 {
     struct sigaction sigact;
 
@@ -103,7 +110,8 @@ static void install_handler (int sig_num, void (*sig_handler) (int sig))
     }
 }
 
-static void print_stats(int iters, const char * str)
+static void
+print_stats(int iters, const char * str)
 {
     if (orig_count > 0) {
         if (0 != dd_count)
@@ -118,7 +126,8 @@ static void print_stats(int iters, const char * str)
         pr2serr("%s commands issued: %d\n", (str ? str : ""), iters);
 }
 
-static void interrupt_handler(int sig)
+static void
+interrupt_handler(int sig)
 {
     struct sigaction sigact;
 
@@ -131,14 +140,16 @@ static void interrupt_handler(int sig)
     kill (getpid (), sig);
 }
 
-static void siginfo_handler(int sig)
+static void
+siginfo_handler(int sig)
 {
     if (sig) { ; }      /* unused, dummy to suppress warning */
     pr2serr("Progress report, continuing ...\n");
     print_stats(0, NULL);
 }
 
-static int dd_filetype(const char * filename)
+static int
+dd_filetype(const char * filename)
 {
     struct stat st;
 
@@ -154,7 +165,8 @@ static int dd_filetype(const char * filename)
     return FT_OTHER;
 }
 
-static void usage()
+static void
+usage()
 {
     pr2serr("Usage: sg_read  [blk_sgio=0|1] [bpt=BPT] [bs=BS] "
             "[cdbsz=6|10|12|16]\n"
@@ -163,7 +175,8 @@ static void usage()
             "                [mmap=0|1] [no_dfxer=0|1] [odir=0|1] "
             "[skip=SKIP]\n"
             "                [time=TI] [verbose=VERB] [--help] "
-            "[--version]\n"
+            "[--verbose]\n"
+            "                [--version] "
             "  where:\n"
             "    blk_sgio 0->normal IO for block devices, 1->SCSI commands "
             "via SG_IO\n"
@@ -194,19 +207,20 @@ static void usage()
             "    time     0->do nothing(def), 1->time from 1st cmd, 2->time "
             "from 2nd, ...\n"
             "    verbose  increase level of verbosity (def: 0)\n"
-            "    --help   print this usage message then exit\n"
-            "    --version  print version number then exit\n\n"
+            "    --help|-h    print this usage message then exit\n"
+            "    --verbose|-v   increase level of verbosity (def: 0)\n"
+            "    --version|-V   print version number then exit\n\n"
             "Issue SCSI READ commands, each starting from the same logical "
             "block address\n");
 }
 
-static int sg_build_scsi_cdb(unsigned char * cdbp, int cdb_sz,
-                             unsigned int blocks, int64_t start_block,
-                             int write_true, int fua, int dpo)
+static int
+sg_build_scsi_cdb(uint8_t * cdbp, int cdb_sz, unsigned int blocks,
+                  int64_t start_block, bool write_true, bool fua, bool dpo)
 {
+    int sz_ind;
     int rd_opcode[] = {0x8, 0x28, 0xa8, 0x88};
     int wr_opcode[] = {0xa, 0x2a, 0xaa, 0x8a};
-    int sz_ind;
 
     memset(cdbp, 0, cdb_sz);
     if (dpo)
@@ -216,10 +230,10 @@ static int sg_build_scsi_cdb(unsigned char * cdbp, int cdb_sz,
     switch (cdb_sz) {
     case 6:
         sz_ind = 0;
-        cdbp[0] = (unsigned char)(write_true ? wr_opcode[sz_ind] :
+        cdbp[0] = (uint8_t)(write_true ? wr_opcode[sz_ind] :
                                                rd_opcode[sz_ind]);
         sg_put_unaligned_be24(0x1fffff & start_block, cdbp + 1);
-        cdbp[4] = (256 == blocks) ? 0 : (unsigned char)blocks;
+        cdbp[4] = (256 == blocks) ? 0 : (uint8_t)blocks;
         if (blocks > 256) {
             pr2serr(ME "for 6 byte commands, maximum number of blocks is "
                     "256\n");
@@ -238,7 +252,7 @@ static int sg_build_scsi_cdb(unsigned char * cdbp, int cdb_sz,
         break;
     case 10:
         sz_ind = 1;
-        cdbp[0] = (unsigned char)(write_true ? wr_opcode[sz_ind] :
+        cdbp[0] = (uint8_t)(write_true ? wr_opcode[sz_ind] :
                                                rd_opcode[sz_ind]);
         sg_put_unaligned_be32((uint32_t)start_block, cdbp + 2);
         sg_put_unaligned_be16((uint16_t)blocks, cdbp + 7);
@@ -250,14 +264,14 @@ static int sg_build_scsi_cdb(unsigned char * cdbp, int cdb_sz,
         break;
     case 12:
         sz_ind = 2;
-        cdbp[0] = (unsigned char)(write_true ? wr_opcode[sz_ind] :
+        cdbp[0] = (uint8_t)(write_true ? wr_opcode[sz_ind] :
                                                rd_opcode[sz_ind]);
         sg_put_unaligned_be32((uint32_t)start_block, cdbp + 2);
         sg_put_unaligned_be32((uint32_t)blocks, cdbp + 6);
         break;
     case 16:
         sz_ind = 3;
-        cdbp[0] = (unsigned char)(write_true ? wr_opcode[sz_ind] :
+        cdbp[0] = (uint8_t)(write_true ? wr_opcode[sz_ind] :
                                                rd_opcode[sz_ind]);
         sg_put_unaligned_be64(start_block, cdbp + 2);
         sg_put_unaligned_be32((uint32_t)blocks, cdbp + 10);
@@ -273,17 +287,18 @@ static int sg_build_scsi_cdb(unsigned char * cdbp, int cdb_sz,
 /* -3 medium/hardware error, -2 -> not ready, 0 -> successful,
    1 -> recoverable (ENOMEM), 2 -> try again (e.g. unit attention),
    3 -> try again (e.g. aborted command), -1 -> other unrecoverable error */
-static int sg_bread(int sg_fd, unsigned char * buff, int blocks,
-                    int64_t from_block, int bs, int cdbsz,
-                    int fua, int dpo, int * diop, int do_mmap,
-                    int no_dxfer)
+static int
+sg_bread(int sg_fd, uint8_t * buff, int blocks, int64_t from_block, int bs,
+         int cdbsz, bool fua, bool dpo, bool * diop, bool do_mmap,
+         bool no_dxfer)
 {
     int k;
-    unsigned char rdCmd[MAX_SCSI_CDBSZ];
-    unsigned char senseBuff[SENSE_BUFF_LEN];
+    uint8_t rdCmd[MAX_SCSI_CDBSZ];
+    uint8_t senseBuff[SENSE_BUFF_LEN];
     struct sg_io_hdr io_hdr;
 
-    if (sg_build_scsi_cdb(rdCmd, cdbsz, blocks, from_block, 0, fua, dpo)) {
+    if (sg_build_scsi_cdb(rdCmd, cdbsz, blocks, from_block, false, fua,
+                          dpo)) {
         pr2serr(ME "bad cdb build, from_block=%" PRId64 ", blocks=%d\n",
                 from_block, blocks);
         return -1;
@@ -331,26 +346,26 @@ static int sg_bread(int sg_fd, unsigned char * buff, int blocks,
         break;
     case SG_LIB_CAT_RECOVERED:
         if (verbose > 1)
-                sg_chk_n_print3("reading, continue", &io_hdr, 1);
+                sg_chk_n_print3("reading, continue", &io_hdr, true);
         break;
     case SG_LIB_CAT_UNIT_ATTENTION:
         if (verbose)
-            sg_chk_n_print3("reading", &io_hdr, verbose - 1);
+            sg_chk_n_print3("reading", &io_hdr, (verbose > 1));
         return 2;
     case SG_LIB_CAT_ABORTED_COMMAND:
         if (verbose)
-            sg_chk_n_print3("reading", &io_hdr, verbose - 1);
+            sg_chk_n_print3("reading", &io_hdr, (verbose > 1));
         return 3;
     case SG_LIB_CAT_NOT_READY:
         if (verbose)
-            sg_chk_n_print3("reading", &io_hdr, verbose - 1);
+            sg_chk_n_print3("reading", &io_hdr, (verbose > 1));
         return -2;
     case SG_LIB_CAT_MEDIUM_HARD:
         if (verbose)
-            sg_chk_n_print3("reading", &io_hdr, verbose - 1);
+            sg_chk_n_print3("reading", &io_hdr, (verbose > 1));
         return -3;
     default:
-        sg_chk_n_print3("reading", &io_hdr, verbose);
+        sg_chk_n_print3("reading", &io_hdr, !! verbose);
         return -1;
     }
     if (blocks > 0) {
@@ -362,42 +377,60 @@ static int sg_bread(int sg_fd, unsigned char * buff, int blocks,
     return 0;
 }
 
+/* Returns the number of times 'ch' is found in string 's' given the
+ * string's length. */
+static int
+num_chs_in_str(const char * s, int slen, int ch)
+{
+    int res = 0;
+
+    while (--slen >= 0) {
+        if (ch == s[slen])
+            ++res;
+    }
+    return res;
+}
+
 #define STR_SZ 1024
 #define INF_SZ 512
-#define EBUFF_SZ 512
+#define EBUFF_SZ 768
 
 
-int main(int argc, char * argv[])
+int
+main(int argc, char * argv[])
 {
-    int64_t skip = 0;
+    bool count_given = false;
+    bool dio_tmp;
+    bool do_blk_sgio = false;
+    bool do_dio = false;
+    bool do_mmap = false;
+    bool do_odir = false;
+    bool dpo = false;
+    bool fua = false;
+    bool no_dxfer = false;
+    bool verbose_given = false;
+    bool version_given = false;
     int bs = 0;
     int bpt = DEF_BLOCKS_PER_TRANSFER;
-    char str[STR_SZ];
+    int dio_incomplete = 0;
+    int do_time = 0;
+    int in_type = FT_OTHER;
+    int ret = 0;
+    int scsi_cdbsz = DEF_SCSI_CDBSZ;
+    int res, k, t, buf_sz, iters, infd, blocks, flags, blocks_per, err;
+    int n, keylen;
+    size_t psz;
+    int64_t skip = 0;
     char * key;
     char * buf;
+    uint8_t * wrkBuff = NULL;
+    uint8_t * wrkPos = NULL;
     char inf[INF_SZ];
     char outf[INF_SZ];
-    int in_type = FT_OTHER;
-    int do_dio = 0;
-    int do_odir = 0;
-    int do_blk_sgio = 0;
-    int do_mmap = 0;
-    int no_dxfer = 0;
-    int do_time = 0;
-    int fua = 0;
-    int dpo = 0;
-    int scsi_cdbsz = DEF_SCSI_CDBSZ;
-    int dio_incomplete = 0;
-    int count_given = 0;
-    int res, k, t, buf_sz, dio_tmp, iters;
-    int infd, blocks, flags, blocks_per;
-    unsigned char * wrkBuff = NULL;
-    unsigned char * wrkPos = NULL;
+    char str[STR_SZ];
     char ebuff[EBUFF_SZ];
-    struct timeval start_tm, end_tm;
     const char * read_str;
-    int ret = 0;
-    size_t psz;
+    struct timeval start_tm, end_tm;
 
 #if defined(HAVE_SYSCONF) && defined(_SC_PAGESIZE)
     psz = sysconf(_SC_PAGESIZE); /* POSIX.1 (was getpagesize()) */
@@ -416,8 +449,9 @@ int main(int argc, char * argv[])
             buf++;
         if (*buf)
             *buf++ = '\0';
+        keylen = strlen(key);
         if (0 == strcmp(key,"blk_sgio"))
-            do_blk_sgio = sg_get_num(buf);
+            do_blk_sgio = !! sg_get_num(buf);
         else if (0 == strcmp(key,"bpt")) {
             bpt = sg_get_num(buf);
             if (-1 == bpt) {
@@ -433,7 +467,7 @@ int main(int argc, char * argv[])
         } else if (0 == strcmp(key,"cdbsz"))
             scsi_cdbsz = sg_get_num(buf);
         else if (0 == strcmp(key,"count")) {
-            count_given = 1;
+            count_given = true;
             if ('-' == *buf) {
                 dd_count = sg_get_llnum(buf + 1);
                 if (-1 == dd_count) {
@@ -449,21 +483,21 @@ int main(int argc, char * argv[])
                 }
             }
         } else if (0 == strcmp(key,"dio"))
-            do_dio = sg_get_num(buf);
+            do_dio = !! sg_get_num(buf);
         else if (0 == strcmp(key,"dpo"))
-            dpo = sg_get_num(buf);
+            dpo = !! sg_get_num(buf);
         else if (0 == strcmp(key,"fua"))
-            fua = sg_get_num(buf);
+            fua = !! sg_get_num(buf);
         else if (strcmp(key,"if") == 0)
-            strncpy(inf, buf, INF_SZ);
+            strncpy(inf, buf, INF_SZ - 1);
         else if (0 == strcmp(key,"mmap"))
-            do_mmap = sg_get_num(buf);
+            do_mmap = !! sg_get_num(buf);
         else if (0 == strcmp(key,"no_dxfer"))
-            no_dxfer = sg_get_num(buf);
+            no_dxfer = !! sg_get_num(buf);
         else if (0 == strcmp(key,"odir"))
-            do_odir = sg_get_num(buf);
+            do_odir = !! sg_get_num(buf);
         else if (strcmp(key,"of") == 0)
-            strncpy(outf, buf, INF_SZ);
+            strncpy(outf, buf, INF_SZ - 1);
         else if (0 == strcmp(key,"skip")) {
             skip = sg_get_llnum(buf);
             if (-1 == skip) {
@@ -472,20 +506,66 @@ int main(int argc, char * argv[])
             }
         } else if (0 == strcmp(key,"time"))
             do_time = sg_get_num(buf);
-        else if (0 == strncmp(key, "verb", 4))
+        else if (0 == strncmp(key, "verb", 4)) {
+            verbose_given = true;
             verbose = sg_get_num(buf);
-        else if (0 == strncmp(key, "--help", 6)) {
+        } else if (0 == strncmp(key, "--help", 6)) {
             usage();
             return 0;
-        } else if (0 == strncmp(key, "--vers", 6)) {
-            pr2serr( ME ": %s\n", version_str);
-            return 0;
+        } else if (0 == strncmp(key, "--verb", 6)) {
+            verbose_given = true;
+            ++verbose;
+        } else if (0 == strncmp(key, "--vers", 6))
+            version_given = true;
+        else if ((keylen > 1) && ('-' == key[0]) && ('-' != key[1])) {
+            res = 0;
+            n = num_chs_in_str(key + 1, keylen - 1, 'h');
+            if (n > 0) {
+                usage();
+                return 0;
+            }
+            n = num_chs_in_str(key + 1, keylen - 1, 'v');
+            if (n > 0)
+                verbose_given = true;
+            verbose += n;
+            res += n;
+            n = num_chs_in_str(key + 1, keylen - 1, 'V');
+            if (n > 0)
+                version_given = true;
+            res += n;
+            if (res < (keylen - 1)) {
+                pr2serr("Unrecognised short option in '%s', try '--help'\n",
+                        key);
+                return SG_LIB_SYNTAX_ERROR;
+            }
         } else {
             pr2serr( "Unrecognized argument '%s'\n", key);
             usage();
             return SG_LIB_SYNTAX_ERROR;
         }
     }
+
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (verbose_given && version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        verbose_given = false;
+        version_given = false;
+        verbose = 0;
+    } else if (! verbose_given) {
+        pr2serr("set '-vv'\n");
+        verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", verbose);
+#else
+    if (verbose_given && version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (version_given) {
+        pr2serr( ME ": %s\n", version_str);
+        return 0;
+    }
+
     if (bs <= 0) {
         bs = DEF_BLOCK_SIZE;
         if ((dd_count > 0) && (bpt > 0))
@@ -511,11 +591,11 @@ int main(int argc, char * argv[])
     }
     if (do_dio && do_mmap) {
         pr2serr("cannot select both dio and mmap\n");
-        return SG_LIB_SYNTAX_ERROR;
+        return SG_LIB_CONTRADICT;
     }
     if (no_dxfer && (do_dio || do_mmap)) {
         pr2serr("cannot select no_dxfer with dio or mmap\n");
-        return SG_LIB_SYNTAX_ERROR;
+        return SG_LIB_CONTRADICT;
     }
 
     install_handler (SIGINT, interrupt_handler);
@@ -553,10 +633,11 @@ int main(int argc, char * argv[])
             if (do_odir)
                 flags |= O_DIRECT;
             if ((infd = open(inf, flags)) < 0) {
+                err = errno;
                 snprintf(ebuff, EBUFF_SZ,
                          ME "could not open %s for sg reading", inf);
                 perror(ebuff);
-                return SG_LIB_FILE_ERROR;
+                return sg_convert_errno(err);
             }
         }
         if (verbose)
@@ -595,10 +676,11 @@ int main(int argc, char * argv[])
         if (do_odir)
             flags |= O_DIRECT;
         if ((infd = open(inf, flags)) < 0) {
+            err = errno;
             snprintf(ebuff,  EBUFF_SZ,
                      ME "could not open %s for reading", inf);
             perror(ebuff);
-            return SG_LIB_FILE_ERROR;
+            return sg_convert_errno(err);
         }
         if (verbose)
             pr2serr("Opened %s for Unix reads with flags=0x%x\n", inf, flags);
@@ -607,10 +689,11 @@ int main(int argc, char * argv[])
 
             offset *= bs;       /* could exceed 32 bits here! */
             if (lseek64(infd, offset, SEEK_SET) < 0) {
+                err = errno;
                 snprintf(ebuff,  EBUFF_SZ,
                     ME "couldn't skip to required position on %s", inf);
                 perror(ebuff);
-                return SG_LIB_FILE_ERROR;
+                return sg_convert_errno(err);
             }
         }
     }
@@ -621,23 +704,23 @@ int main(int argc, char * argv[])
 
     if (dd_count > 0) {
         if (do_dio || do_odir || (FT_RAW & in_type)) {
-            wrkBuff = (unsigned char *)malloc(bs * bpt + psz);
+            wrkBuff = (uint8_t *)malloc(bs * bpt + psz);
             if (0 == wrkBuff) {
                 pr2serr("Not enough user memory for aligned storage\n");
                 return SG_LIB_CAT_OTHER;
             }
             /* perhaps use posix_memalign() instead */
-            wrkPos = (unsigned char *)(((uintptr_t)wrkBuff + psz - 1) &
+            wrkPos = (uint8_t *)(((sg_uintptr_t)wrkBuff + psz - 1) &
                                        (~(psz - 1)));
         } else if (do_mmap) {
-            wrkPos = (unsigned char *)mmap(NULL, bs * bpt,
+            wrkPos = (uint8_t *)mmap(NULL, bs * bpt,
                         PROT_READ | PROT_WRITE, MAP_SHARED, infd, 0);
             if (MAP_FAILED == wrkPos) {
                 perror(ME "error from mmap()");
                 return SG_LIB_CAT_OTHER;
             }
         } else {
-            wrkBuff = (unsigned char *)malloc(bs * bpt);
+            wrkBuff = (uint8_t *)malloc(bs * bpt);
             if (0 == wrkBuff) {
                 pr2serr("Not enough user memory\n");
                 return SG_LIB_CAT_OTHER;
@@ -809,7 +892,6 @@ int main(int argc, char * argv[])
         free(wrkBuff);
 
     close(infd);
-    res = 0;
     if (0 != dd_count) {
         pr2serr("Some error occurred,");
         if (0 == ret)
