@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2006-2016 Douglas Gilbert.
+ * Copyright (c) 2006-2018 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
@@ -9,6 +9,9 @@
 #include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
+#include <stdbool.h>
+#include <errno.h>
 #include <string.h>
 #include <getopt.h>
 #define __STDC_FORMAT_MACROS 1
@@ -17,12 +20,13 @@
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
+
 #include "sg_lib.h"
 #include "sg_cmds_basic.h"
 #include "sg_cmds_extra.h"
 #include "sg_pr2serr.h"
 
-static const char * version_str = "1.07 20160126";
+static const char * version_str = "1.13 20180628";
 
 /* This program uses a ATA PASS-THROUGH SCSI command. This usage is
  * defined in the SCSI to ATA Translation (SAT) drafts and standards.
@@ -44,7 +48,7 @@ static const char * version_str = "1.07 20160126";
 
 #define SAT_ATA_PASS_THROUGH16 0x85
 #define SAT_ATA_PASS_THROUGH16_LEN 16
-#define SAT_ATA_PASS_THROUGH12 0xa1     /* clashes with MMC BLANK comand */
+#define SAT_ATA_PASS_THROUGH12 0xa1     /* clashes with MMC BLANK command */
 #define SAT_ATA_PASS_THROUGH12_LEN 12
 #define SAT_ATA_RETURN_DESC 9  /* ATA Return (sense) Descriptor */
 #define ASCQ_ATA_PT_INFO_AVAILABLE 0x1d
@@ -59,12 +63,14 @@ static const char * version_str = "1.07 20160126";
 
 static struct option long_options[] = {
         {"ck_cond", no_argument, 0, 'c'},
+        {"ck-cond", no_argument, 0, 'c'},
         {"extend", no_argument, 0, 'e'},
         {"hex", no_argument, 0, 'H'},
         {"ignore", no_argument, 0, 'i'},
         {"len", no_argument, 0, 'l'},
         {"raw", no_argument, 0, 'r'},
         {"reset", no_argument, 0, 'R'},
+        {"help", no_argument, 0, 'h'},
         {"verbose", no_argument, 0, 'v'},
         {"version", no_argument, 0, 'V'},
         {0, 0, 0, 0},
@@ -137,81 +143,100 @@ find_phy_desc(int id)
 }
 
 static void
-dStrRaw(const char* str, int len)
+dStrRaw(const uint8_t * str, int len)
 {
     int k;
 
-    for (k = 0 ; k < len; ++k)
+    for (k =0; k < len; ++k)
         printf("%c", str[k]);
 }
 
 /* ATA READ LOG EXT command [2Fh, PIO data-in] */
-/* N.B. "log_addr" is the log page number, "page_in_log" is usually zero */
+/* N.B. "log_addr" is the log page number, "page_in_log" is usually false */
 static int
-do_read_log_ext(int sg_fd, int log_addr, int page_in_log, int feature,
+do_read_log_ext(int sg_fd, int log_addr, bool page_in_log, int feature,
                 int blk_count, void * resp, int mx_resp_len, int cdb_len,
-                int ck_cond, int extend, int do_hex, int do_raw, int verbose)
+                bool ck_cond, bool extend, int do_hex, bool do_raw,
+                int verbose)
 {
-    int ok, res, ret;
     /* Following for ATA READ/WRITE MULTIPLE (EXT) cmds, normally 0 */
+    bool t_type = false;/* false -> 512 byte LBs, true -> device's LB size */
+    bool t_dir = true;  /* false -> to device, 1 -> from device */
+    bool byte_block = true; /* false -> bytes, true -> 512 byte blocks (if
+                               t_type=false) */
+    bool got_ard = false;    /* got ATA result descriptor */
+    bool ok;
+    int res, ret;
     int multiple_count = 0;
     int protocol = 4;   /* PIO data-in */
-    int t_type = 0;     /* 0 -> 512 byte blocks, 1 -> device's LB size */
-    int t_dir = 1;      /* 0 -> to device, 1 -> from device */
-    int byte_block = 1; /* 0 -> bytes, 1 -> 512 byte blocks (if t_type=0) */
     int t_length = 2;   /* 0 -> no data transferred, 2 -> sector count */
     int resid = 0;
-    int got_ard = 0;    /* got ATA result descriptor */
     int sb_sz;
     struct sg_scsi_sense_hdr ssh;
-    unsigned char sense_buffer[64];
-    unsigned char ata_return_desc[16];
-    unsigned char aptCmdBlk[SAT_ATA_PASS_THROUGH16_LEN] =
+    uint8_t sense_buffer[64];
+    uint8_t ata_return_desc[16];
+    uint8_t apt_cdb[SAT_ATA_PASS_THROUGH16_LEN] =
                 {SAT_ATA_PASS_THROUGH16, 0, 0, 0, 0, 0, 0, 0,
                  0, 0, 0, 0, 0, 0, 0, 0};
-    unsigned char apt12CmdBlk[SAT_ATA_PASS_THROUGH12_LEN] =
+    uint8_t apt12_cdb[SAT_ATA_PASS_THROUGH12_LEN] =
                 {SAT_ATA_PASS_THROUGH12, 0, 0, 0, 0, 0, 0, 0,
                  0, 0, 0, 0};
 
     sb_sz = sizeof(sense_buffer);
     memset(sense_buffer, 0, sb_sz);
     memset(ata_return_desc, 0, sizeof(ata_return_desc));
-    ok = 0;
+    ok = false;
     if (SAT_ATA_PASS_THROUGH16_LEN == cdb_len) {
         /* Prepare ATA PASS-THROUGH COMMAND (16) command */
-        aptCmdBlk[3] = (feature >> 8) & 0xff;   /* feature(15:8) */
-        aptCmdBlk[4] = feature & 0xff;          /* feature(7:0) */
-        aptCmdBlk[5] = (blk_count >> 8) & 0xff; /* sector_count(15:8) */
-        aptCmdBlk[6] = blk_count & 0xff;        /* sector_count(7:0) */
-        aptCmdBlk[8] = log_addr & 0xff;  /* lba_low(7:0) == LBA(7:0) */
-        aptCmdBlk[9] = (page_in_log >> 8) & 0xff;
+        apt_cdb[3] = (feature >> 8) & 0xff;   /* feature(15:8) */
+        apt_cdb[4] = feature & 0xff;          /* feature(7:0) */
+        apt_cdb[5] = (blk_count >> 8) & 0xff; /* sector_count(15:8) */
+        apt_cdb[6] = blk_count & 0xff;        /* sector_count(7:0) */
+        apt_cdb[8] = log_addr & 0xff;  /* lba_low(7:0) == LBA(7:0) */
+        apt_cdb[9] = (page_in_log >> 8) & 0xff;
                 /* lba_mid(15:8) == LBA(39:32) */
-        aptCmdBlk[10] = page_in_log & 0xff; /* lba_mid(7:0) == LBA(15:8) */
-        aptCmdBlk[14] = ATA_READ_LOG_EXT;
-        aptCmdBlk[1] = (multiple_count << 5) | (protocol << 1) | extend;
-        aptCmdBlk[2] = (ck_cond << 5) | (t_type << 4) | (t_dir << 3) |
-                       (byte_block << 2) | t_length;
-        res = sg_ll_ata_pt(sg_fd, aptCmdBlk, cdb_len, DEF_TIMEOUT, resp,
+        apt_cdb[10] = page_in_log & 0xff; /* lba_mid(7:0) == LBA(15:8) */
+        apt_cdb[14] = ATA_READ_LOG_EXT;
+        apt_cdb[1] = (multiple_count << 5) | (protocol << 1);
+        if (extend)
+            apt_cdb[1] |= 0x1;
+        apt_cdb[2] = t_length;
+        if (ck_cond)
+            apt_cdb[2] |= 0x20;
+        if (t_type)
+            apt_cdb[2] |= 0x10;
+        if (t_dir)
+            apt_cdb[2] |= 0x8;
+        if (byte_block)
+            apt_cdb[2] |= 0x4;
+        res = sg_ll_ata_pt(sg_fd, apt_cdb, cdb_len, DEF_TIMEOUT, resp,
                            NULL /* doutp */, mx_resp_len, sense_buffer,
                            sb_sz, ata_return_desc,
                            sizeof(ata_return_desc), &resid, verbose);
     } else {
         /* Prepare ATA PASS-THROUGH COMMAND (12) command */
-        apt12CmdBlk[3] = feature & 0xff;        /* feature(7:0) */
-        apt12CmdBlk[4] = blk_count & 0xff;        /* sector_count(7:0) */
-        apt12CmdBlk[5] = log_addr & 0xff;  /* lba_low(7:0) == LBA(7:0) */
-        apt12CmdBlk[6] = page_in_log & 0xff; /* lba_mid(7:0) == LBA(15:8) */
-        apt12CmdBlk[9] = ATA_READ_LOG_EXT;
-        apt12CmdBlk[1] = (multiple_count << 5) | (protocol << 1);
-        apt12CmdBlk[2] = (ck_cond << 5) | (t_type << 4) | (t_dir << 3) |
-                         (byte_block << 2) | t_length;
-        res = sg_ll_ata_pt(sg_fd, apt12CmdBlk, cdb_len, DEF_TIMEOUT, resp,
+        apt12_cdb[3] = feature & 0xff;        /* feature(7:0) */
+        apt12_cdb[4] = blk_count & 0xff;        /* sector_count(7:0) */
+        apt12_cdb[5] = log_addr & 0xff;  /* lba_low(7:0) == LBA(7:0) */
+        apt12_cdb[6] = page_in_log & 0xff; /* lba_mid(7:0) == LBA(15:8) */
+        apt12_cdb[9] = ATA_READ_LOG_EXT;
+        apt12_cdb[1] = (multiple_count << 5) | (protocol << 1);
+        apt12_cdb[2] = t_length;
+        if (ck_cond)
+            apt12_cdb[2] |= 0x20;
+        if (t_type)
+            apt12_cdb[2] |= 0x10;
+        if (t_dir)
+            apt12_cdb[2] |= 0x8;
+        if (byte_block)
+            apt12_cdb[2] |= 0x4;
+        res = sg_ll_ata_pt(sg_fd, apt12_cdb, cdb_len, DEF_TIMEOUT, resp,
                            NULL /* doutp */, mx_resp_len, sense_buffer,
                            sb_sz, ata_return_desc,
                            sizeof(ata_return_desc), &resid, verbose);
     }
     if (0 == res) {
-        ok = 1;
+        ok = true;
         if (verbose > 2)
             pr2serr("command completed with SCSI GOOD status\n");
     } else if ((res > 0) && (res & SAM_STAT_CHECK_CONDITION)) {
@@ -245,7 +270,7 @@ do_read_log_ext(int sg_fd, int log_addr, int page_in_log, int feature,
                                     "Descriptor\n");
                         return SG_LIB_CAT_RECOVERED;
                     }
-                    got_ard = 1;
+                    got_ard = true;
                     break;
                 } else if (SPC_SK_RECOVERED_ERROR == ssh.sense_key)
                     return SG_LIB_CAT_RECOVERED;
@@ -312,7 +337,7 @@ do_read_log_ext(int sg_fd, int log_addr, int page_in_log, int feature,
         return -1;
     }
 
-    if ((SAT_ATA_RETURN_DESC == ata_return_desc[0]) && (0 == got_ard))
+    if ((SAT_ATA_RETURN_DESC == ata_return_desc[0]) && (! got_ard))
         pr2serr("Seem to have got ATA Result Descriptor but it was not "
                 "indicated\n");
     if (got_ard) {
@@ -320,14 +345,14 @@ do_read_log_ext(int sg_fd, int log_addr, int page_in_log, int feature,
                 pr2serr("error indication in returned FIS: aborted command\n");
                 return SG_LIB_CAT_ABORTED_COMMAND;
         }
-        ok = 1;
+        ok = true;
     }
 
     if (ok) { /* output result if ok and --hex or --raw given */
         if (do_raw)
-            dStrRaw((const char *)resp, mx_resp_len);
+            dStrRaw((const uint8_t *)resp, mx_resp_len);
         else if (1 == do_hex)
-            dStrHex((const char *)resp, mx_resp_len, 0);
+            hex2stdout((const uint8_t *)resp, mx_resp_len, 0);
         else if (do_hex > 1)
             dWordHex((const unsigned short *)resp, mx_resp_len / 2, 0,
                      sg_is_big_endian());
@@ -338,18 +363,20 @@ do_read_log_ext(int sg_fd, int log_addr, int page_in_log, int feature,
 
 int main(int argc, char * argv[])
 {
-    int sg_fd, c, k, j, res, id, len, vendor;
+    bool ck_cond = false;   /* set to true to read register(s) back */
+    bool extend = false;
+    bool ignore = false;
+    bool raw = false;
+    bool reset = false;
+    bool verbose_given = false;
+    bool version_given = false;
+    int sg_fd, c, k, j, res, id, len, vendor, err;
     char * device_name = 0;
     char ebuff[EBUFF_SZ];
-    unsigned char inBuff[READ_LOG_EXT_RESPONSE_LEN];
+    uint8_t inBuff[READ_LOG_EXT_RESPONSE_LEN];
     int cdb_len = 16;
     int hex = 0;
-    int ignore = 0;
-    int raw = 0;
-    int reset = 0;
     int verbose = 0;
-    int ck_cond = 0;   /* set to 1 to read register(s) back */
-    int extend = 0;
     int ret = 0;
     uint64_t ull;
     const char * cp;
@@ -365,10 +392,10 @@ int main(int argc, char * argv[])
 
         switch (c) {
         case 'c':
-            ++ck_cond;
+            ck_cond = true;
             break;
         case 'e':
-            ++extend;
+            extend = true;
             break;
         case 'h':
         case '?':
@@ -378,7 +405,7 @@ int main(int argc, char * argv[])
             ++hex;
             break;
         case 'i':
-            ++ignore;
+            ignore = true;
             break;
         case 'l':
             cdb_len = sg_get_num(optarg);
@@ -388,17 +415,18 @@ int main(int argc, char * argv[])
             }
             break;
         case 'r':
-            ++raw;
+            raw = true;
             break;
         case 'R':
-            ++reset;
+            reset = true;
             break;
         case 'v':
+            verbose_given = true;
             ++verbose;
             break;
         case 'V':
-            pr2serr("version: %s\n", version_str);
-            exit(0);
+            version_given = true;
+            break;
         default:
             pr2serr("unrecognised option code %c [0x%x]\n", c, c);
             usage();
@@ -417,8 +445,28 @@ int main(int argc, char * argv[])
             return SG_LIB_SYNTAX_ERROR;
         }
     }
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (verbose_given && version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        verbose_given = false;
+        version_given = false;
+        verbose = 0;
+    } else if (! verbose_given) {
+        pr2serr("set '-vv'\n");
+        verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", verbose);
+#else
+    if (verbose_given && version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (version_given) {
+        pr2serr("version: %s\n", version_str);
+        return 0;
+    }
     if (0 == device_name) {
-        pr2serr("no DEVICE name detected\n");
+        pr2serr("no DEVICE name detected\n\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
@@ -430,18 +478,20 @@ int main(int argc, char * argv[])
     }
 
     if ((sg_fd = open(device_name, O_RDWR)) < 0) {
+        err = errno;
         snprintf(ebuff, EBUFF_SZ,
                  "sg_sat_phy_event: error opening file: %s", device_name);
         perror(ebuff);
-        return SG_LIB_FILE_ERROR;
+        return sg_convert_errno(err);
     }
-    ret = do_read_log_ext(sg_fd, SATA_PHY_EVENT_LPAGE, 0 /* page_in_log */,
+    ret = do_read_log_ext(sg_fd, SATA_PHY_EVENT_LPAGE,
+                          false /* page_in_log */,
                           (reset ? 1 : 0) /* feature */,
                           1 /* blk_count */, inBuff,
                           READ_LOG_EXT_RESPONSE_LEN, cdb_len, ck_cond,
                           extend, hex, raw, verbose);
 
-    if ((0 == ret) && (0 == hex) && (0 == raw)) {
+    if ((0 == ret) && (0 == hex) && (! raw)) {
         printf("SATA phy event counters:\n");
         for (k = 4; k < 512; k += (len + 2)) {
             id = (inBuff[k + 1] << 8) + inBuff[k];
@@ -457,7 +507,7 @@ int main(int argc, char * argv[])
                 ull |= inBuff[k + 2 + j];
             }
             cp = NULL;
-            if ((0 == vendor) && (0 == ignore))
+            if ((0 == vendor) && (! ignore))
                 cp = find_phy_desc(id);
             if (cp)
                 printf("  %s: %" PRIu64 "\n", cp, ull);
@@ -469,9 +519,10 @@ int main(int argc, char * argv[])
 
     res = close(sg_fd);
     if (res < 0) {
-        pr2serr("close error: %s\n", safe_strerror(-res));
+        err = errno;
+        pr2serr("close error: %s\n", safe_strerror(err));
         if (0 == ret)
-            return SG_LIB_FILE_ERROR;
+            ret = sg_convert_errno(err);
     }
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
