@@ -1,9 +1,9 @@
 /*
  * (c) 2000 Kurt Garloff <garloff at suse dot de>
  * heavily based on Douglas Gilbert's sg_rbuf program.
- * (c) 1999-2015 Douglas Gilbert
+ * (c) 1999-2018 Douglas Gilbert
  *
- * Program to test the SCSI host adapter by issueing
+ * Program to test the SCSI host adapter by issuing
  * write and read operations on a device's buffer
  * and calculating checksums.
  * NOTE: If you can not reserve the buffer of the device
@@ -26,7 +26,10 @@
 #include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
+#include <stdbool.h>
 #include <string.h>
+#include <errno.h>
 #include <getopt.h>
 #include <sys/ioctl.h>
 #include <sys/types.h>
@@ -35,13 +38,14 @@
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
+
 #include "sg_lib.h"
 #include "sg_io_linux.h"
 #include "sg_unaligned.h"
 #include "sg_pr2serr.h"
 
 
-static const char * version_str = "1.09 20151220";
+static const char * version_str = "1.18 20180628";
 
 #define BPI (signed)(sizeof(int))
 
@@ -60,53 +64,55 @@ static const char * version_str = "1.09 20151220";
 static int base = 0x12345678;
 static int buf_capacity = 0;
 static int buf_granul = 255;
-static unsigned char *cmpbuf = NULL;
+static uint8_t *cmpbuf = NULL;
+static uint8_t *free_cmpbuf = NULL;
 
 
 /* Options */
 static int size = -1;
-static char do_quick = 0;
+static bool do_quick = false;
 static int addwrite  = 0;
 static int addread   = 0;
 static int verbose   = 0;
 
 static struct option long_options[] = {
-        {"help", 0, 0, 'h'},
-        {"quick", 0, 0, 'q'},
-        {"addrd", 1, 0, 'r'},
-        {"size", 1, 0, 's'},
-        {"times", 1, 0, 't'},
-        {"verbose", 0, 0, 'v'},
-        {"version", 0, 0, 'V'},
-        {"addwr", 1, 0, 'w'},
+        {"help", no_argument, 0, 'h'},
+        {"quick", no_argument, 0, 'q'},
+        {"addrd", required_argument, 0, 'r'},
+        {"size", required_argument, 0, 's'},
+        {"times", required_argument, 0, 't'},
+        {"verbose", no_argument, 0, 'v'},
+        {"version", no_argument, 0, 'V'},
+        {"addwr", required_argument, 0, 'w'},
         {0, 0, 0, 0},
 };
 
-int find_out_about_buffer (int sg_fd)
+static int
+find_out_about_buffer(int sg_fd)
 {
-        unsigned char rbCmdBlk[] = {READ_BUFFER, 0, 0, 0, 0, 0, 0, 0, 0, 0};
-        unsigned char rbBuff[RB_DESC_LEN];
-        unsigned char sense_buffer[32];
+        uint8_t rb_cdb[] = {READ_BUFFER, 0, 0, 0, 0, 0, 0, 0, 0, 0};
+        uint8_t rbBuff[RB_DESC_LEN];
+        uint8_t sense_buffer[32];
         struct sg_io_hdr io_hdr;
         int k, res;
 
-        rbCmdBlk[1] = RB_MODE_DESC;
-        rbCmdBlk[8] = RB_DESC_LEN;
+        rb_cdb[1] = RB_MODE_DESC;
+        rb_cdb[8] = RB_DESC_LEN;
         memset(&io_hdr, 0, sizeof(struct sg_io_hdr));
         io_hdr.interface_id = 'S';
-        io_hdr.cmd_len = sizeof(rbCmdBlk);
+        io_hdr.cmd_len = sizeof(rb_cdb);
         io_hdr.mx_sb_len = sizeof(sense_buffer);
         io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
         io_hdr.dxfer_len = RB_DESC_LEN;
         io_hdr.dxferp = rbBuff;
-        io_hdr.cmdp = rbCmdBlk;
+        io_hdr.cmdp = rb_cdb;
         io_hdr.sbp = sense_buffer;
         io_hdr.timeout = 60000;     /* 60000 millisecs == 60 seconds */
 
         if (verbose) {
                 pr2serr("    read buffer [mode desc] cdb: ");
-                for (k = 0; k < (int)sizeof(rbCmdBlk); ++k)
-                        pr2serr("%02x ", rbCmdBlk[k]);
+                for (k = 0; k < (int)sizeof(rb_cdb); ++k)
+                        pr2serr("%02x ", rb_cdb[k]);
                 pr2serr("\n");
         }
         if (ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
@@ -118,17 +124,23 @@ int find_out_about_buffer (int sg_fd)
         switch (res) {
         case SG_LIB_CAT_RECOVERED:
                 sg_chk_n_print3("READ BUFFER descriptor, continuing",
-                                &io_hdr, 1);
-                /* fall through */
+                                &io_hdr, true);
+#if defined(__GNUC__)
+#if (__GNUC__ >= 7)
+                __attribute__((fallthrough));
+                /* FALL THROUGH */
+#endif
+#endif
         case SG_LIB_CAT_CLEAN:
                 break;
         default: /* won't bother decoding other categories */
-                sg_chk_n_print3("READ BUFFER descriptor error", &io_hdr, 1);
+                sg_chk_n_print3("READ BUFFER descriptor error", &io_hdr,
+                                true);
                 return res;
         }
 
         buf_capacity = sg_get_unaligned_be24(rbBuff + 1);
-        buf_granul = (unsigned char)rbBuff[0];
+        buf_granul = (uint8_t)rbBuff[0];
 #if 0
         printf("READ BUFFER reports: %02x %02x %02x %02x\n",
                rbBuff[0], rbBuff[1], rbBuff[2], rbBuff[3]);
@@ -139,7 +151,8 @@ int find_out_about_buffer (int sg_fd)
         return 0;
 }
 
-int mymemcmp (unsigned char *bf1, unsigned char *bf2, int len)
+static int
+mymemcmp (uint8_t *bf1, uint8_t *bf2, int len)
 {
         int df;
         for (df = 0; df < len; df++)
@@ -148,7 +161,8 @@ int mymemcmp (unsigned char *bf1, unsigned char *bf2, int len)
 }
 
 /* return 0 if good, else 2222 */
-int do_checksum (int *buf, int len, int quiet)
+static int
+do_checksum(int *buf, int len, bool quiet)
 {
         int sum = base;
         int i; int rln = len;
@@ -159,7 +173,7 @@ int do_checksum (int *buf, int len, int quiet)
                 if (!quiet) printf ("sg_test_rwbuf: Checksum error (sz=%i):"
                                     " %08x\n", len, sum);
                 if (cmpbuf && !quiet) {
-                        int diff = mymemcmp (cmpbuf, (unsigned char*)buf,
+                        int diff = mymemcmp (cmpbuf, (uint8_t*)buf,
                                              len);
                         printf ("Differ at pos %i/%i:\n", diff, len);
                         for (i = 0; i < 24 && i+diff < len; i++)
@@ -167,7 +181,7 @@ int do_checksum (int *buf, int len, int quiet)
                         printf ("\n");
                         for (i = 0; i < 24 && i+diff < len; i++)
                                 printf (" %02x",
-                                        ((unsigned char*)buf)[i+diff]);
+                                        ((uint8_t*)buf)[i+diff]);
                         printf ("\n");
                 }
                 return 2222;
@@ -203,7 +217,7 @@ void do_fill_buffer (int *buf, int len)
         }
         if (len >= BPI) buf[len/BPI - 1] = 0x12345678 - sum;
         else ((char*)buf)[0] = 0x12345678 + ((char*)buf)[0] - sum;
-        if (do_checksum (buf, len, 1)) {
+        if (do_checksum(buf, len, true)) {
                 if (len < BPI) goto retry;
                 printf ("sg_test_rwbuf: Memory corruption?\n");
                 exit (1);
@@ -212,34 +226,36 @@ void do_fill_buffer (int *buf, int len)
 }
 
 
-int read_buffer (int sg_fd, unsigned size)
+int read_buffer (int sg_fd, unsigned ssize)
 {
         int res, k;
-        unsigned char rbCmdBlk[] = {READ_BUFFER, 0, 0, 0, 0, 0, 0, 0, 0, 0};
-        int bufSize = size + addread;
-        unsigned char * rbBuff = (unsigned char *)malloc(bufSize);
-        unsigned char sense_buffer[32];
+        uint8_t rb_cdb[] = {READ_BUFFER, 0, 0, 0, 0, 0, 0, 0, 0, 0};
+        int bufSize = ssize + addread;
+        uint8_t * free_rbBuff = NULL;
+        uint8_t * rbBuff = (uint8_t *)sg_memalign(bufSize, 0, &free_rbBuff,
+                                                  false);
+        uint8_t sense_buffer[32];
         struct sg_io_hdr io_hdr;
 
         if (NULL == rbBuff)
                 return -1;
-        rbCmdBlk[1] = RWB_MODE_DATA;
-        sg_put_unaligned_be24((uint32_t)bufSize, rbCmdBlk + 6);
+        rb_cdb[1] = RWB_MODE_DATA;
+        sg_put_unaligned_be24((uint32_t)bufSize, rb_cdb + 6);
         memset(&io_hdr, 0, sizeof(struct sg_io_hdr));
         io_hdr.interface_id = 'S';
-        io_hdr.cmd_len = sizeof(rbCmdBlk);
+        io_hdr.cmd_len = sizeof(rb_cdb);
         io_hdr.mx_sb_len = sizeof(sense_buffer);
         io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
         io_hdr.dxfer_len = bufSize;
         io_hdr.dxferp = rbBuff;
-        io_hdr.cmdp = rbCmdBlk;
+        io_hdr.cmdp = rb_cdb;
         io_hdr.sbp = sense_buffer;
         io_hdr.pack_id = 2;
         io_hdr.timeout = 60000;     /* 60000 millisecs == 60 seconds */
         if (verbose) {
                 pr2serr("    read buffer [mode data] cdb: ");
-                for (k = 0; k < (int)sizeof(rbCmdBlk); ++k)
-                        pr2serr("%02x ", rbCmdBlk[k]);
+                for (k = 0; k < (int)sizeof(rb_cdb); ++k)
+                        pr2serr("%02x ", rb_cdb[k]);
                 pr2serr("\n");
         }
 
@@ -252,51 +268,59 @@ int read_buffer (int sg_fd, unsigned size)
         res = sg_err_category3(&io_hdr);
         switch (res) {
         case SG_LIB_CAT_RECOVERED:
-            sg_chk_n_print3("READ BUFFER data, continuing", &io_hdr, 1);
-            /* fall through */
+            sg_chk_n_print3("READ BUFFER data, continuing", &io_hdr, true);
+#if defined(__GNUC__)
+#if (__GNUC__ >= 7)
+            __attribute__((fallthrough));
+            /* FALL THROUGH */
+#endif
+#endif
         case SG_LIB_CAT_CLEAN:
                 break;
         default: /* won't bother decoding other categories */
-                sg_chk_n_print3("READ BUFFER data error", &io_hdr, 1);
+                sg_chk_n_print3("READ BUFFER data error", &io_hdr, true);
                 free(rbBuff);
                 return res;
         }
 
-        res = do_checksum ((int*)rbBuff, size, 0);
-        free(rbBuff);
+        res = do_checksum((int*)rbBuff, ssize, false);
+        if (free_rbBuff)
+                free(free_rbBuff);
         return res;
 }
 
-int write_buffer (int sg_fd, unsigned size)
+int write_buffer (int sg_fd, unsigned ssize)
 {
-        unsigned char wbCmdBlk[] = {WRITE_BUFFER, 0, 0, 0, 0, 0, 0, 0, 0, 0};
-        int bufSize = size + addwrite;
-        unsigned char * wbBuff = (unsigned char *)malloc(bufSize);
-        unsigned char sense_buffer[32];
+        uint8_t wb_cdb[] = {WRITE_BUFFER, 0, 0, 0, 0, 0, 0, 0, 0, 0};
+        int bufSize = ssize + addwrite;
+        uint8_t * free_wbBuff = NULL;
+        uint8_t * wbBuff = (uint8_t *)sg_memalign(bufSize, 0, &free_wbBuff,
+                                                  false);
+        uint8_t sense_buffer[32];
         struct sg_io_hdr io_hdr;
         int k, res;
 
         if (NULL == wbBuff)
                 return -1;
         memset(wbBuff, 0, bufSize);
-        do_fill_buffer ((int*)wbBuff, size);
-        wbCmdBlk[1] = RWB_MODE_DATA;
-        sg_put_unaligned_be24((uint32_t)bufSize, wbCmdBlk + 6);
+        do_fill_buffer ((int*)wbBuff, ssize);
+        wb_cdb[1] = RWB_MODE_DATA;
+        sg_put_unaligned_be24((uint32_t)bufSize, wb_cdb + 6);
         memset(&io_hdr, 0, sizeof(struct sg_io_hdr));
         io_hdr.interface_id = 'S';
-        io_hdr.cmd_len = sizeof(wbCmdBlk);
+        io_hdr.cmd_len = sizeof(wb_cdb);
         io_hdr.mx_sb_len = sizeof(sense_buffer);
         io_hdr.dxfer_direction = SG_DXFER_TO_DEV;
         io_hdr.dxfer_len = bufSize;
         io_hdr.dxferp = wbBuff;
-        io_hdr.cmdp = wbCmdBlk;
+        io_hdr.cmdp = wb_cdb;
         io_hdr.sbp = sense_buffer;
         io_hdr.pack_id = 1;
         io_hdr.timeout = 60000;     /* 60000 millisecs == 60 seconds */
         if (verbose) {
                 pr2serr("    write buffer [mode data] cdb: ");
-                for (k = 0; k < (int)sizeof(wbCmdBlk); ++k)
-                        pr2serr("%02x ", wbCmdBlk[k]);
+                for (k = 0; k < (int)sizeof(wb_cdb); ++k)
+                        pr2serr("%02x ", wb_cdb[k]);
                 pr2serr("\n");
         }
 
@@ -309,16 +333,22 @@ int write_buffer (int sg_fd, unsigned size)
         res = sg_err_category3(&io_hdr);
         switch (res) {
         case SG_LIB_CAT_RECOVERED:
-            sg_chk_n_print3("WRITE BUFFER data, continuing", &io_hdr, 1);
-            /* fall through */
+            sg_chk_n_print3("WRITE BUFFER data, continuing", &io_hdr, true);
+#if defined(__GNUC__)
+#if (__GNUC__ >= 7)
+            __attribute__((fallthrough));
+            /* FALL THROUGH */
+#endif
+#endif
         case SG_LIB_CAT_CLEAN:
                 break;
         default: /* won't bother decoding other categories */
-                sg_chk_n_print3("WRITE BUFFER data error", &io_hdr, 1);
+                sg_chk_n_print3("WRITE BUFFER data error", &io_hdr, true);
                 free(wbBuff);
                 return res;
         }
-        free(wbBuff);
+        if (free_wbBuff)
+                free(free_wbBuff);
         return res;
 }
 
@@ -332,17 +362,17 @@ void usage ()
                 " or\n"
                 "       sg_test_rwbuf DEVICE SZ [AW] [AR]\n");
         printf ("  where:\n"
-                "    --addrd=AR       extra bytes to fetch during READ "
+                "    --addrd=AR|-r    extra bytes to fetch during READ "
                 "BUFFER\n"
-                "    --addwr=AW       extra bytes to send to WRITE BUFFER\n"
-                "    --help           output this usage message then exit\n"
-                "    --quick          output read buffer size then exit\n"
-                "    --size=SZ        size of buffer (in bytes) to write "
+                "    --addwr=AW|-w    extra bytes to send to WRITE BUFFER\n"
+                "    --help|-l        output this usage message then exit\n"
+                "    --quick|-q       output read buffer size then exit\n"
+                "    --size=SZ|-s     size of buffer (in bytes) to write "
                 "then read back\n"
-                "    --times=NUM      number of times to run test "
+                "    --times=NUM|-t   number of times to run test "
                 "(default 1)\n"
-                "    --verbose        increase verbosity of output\n"
-                "    --version        output version then exit\n");
+                "    --verbose|-v     increase verbosity of output\n"
+                "    --version|-V     output version then exit\n");
         printf ("\nWARNING: If you access the device at the same time, e.g. "
                 "because it's a\n");
         printf (" mounted hard disk, the device's buffer may be used by the "
@@ -356,11 +386,14 @@ void usage ()
 
 int main (int argc, char * argv[])
 {
+        bool verbose_given = true;
+        bool version_given = true;
         int sg_fd, res;
         const char * device_name = NULL;
         int times = 1;
         int ret = 0;
         int k = 0;
+        int err;
 
         while (1) {
                 int option_index = 0;
@@ -376,7 +409,7 @@ int main (int argc, char * argv[])
                         usage();
                         return 0;
                 case 'q':
-                        do_quick = 1;
+                        do_quick = true;
                         break;
                 case 'r':
                         addread = sg_get_num(optarg);
@@ -400,11 +433,12 @@ int main (int argc, char * argv[])
                         }
                         break;
                 case 'v':
+                        verbose_given = true;
                         verbose++;
                         break;
                 case 'V':
-                        pr2serr(ME "version: %s\n", version_str);
-                        return 0;
+                        version_given = true;
+                        break;
                 case 'w':
                         addwrite = sg_get_num(optarg);
                         if (-1 == addwrite) {
@@ -457,6 +491,29 @@ int main (int argc, char * argv[])
                         return SG_LIB_SYNTAX_ERROR;
                 }
         }
+
+#ifdef DEBUG
+        pr2serr("In DEBUG mode, ");
+        if (verbose_given && version_given) {
+                pr2serr("but override: '-vV' given, zero verbose and "
+                        "continue\n");
+                verbose_given = false;
+                version_given = false;
+                verbose = 0;
+        } else if (! verbose_given) {
+                pr2serr("set '-vv'\n");
+                verbose = 2;
+        } else
+                pr2serr("keep verbose=%d\n", verbose);
+#else
+        if (verbose_given && version_given)
+                pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+        if (version_given) {
+                pr2serr(ME "version: %s\n", version_str);
+                return 0;
+        }
+
         if (NULL == device_name) {
                 pr2serr("no device name given\n");
                 usage();
@@ -471,8 +528,9 @@ int main (int argc, char * argv[])
 
         sg_fd = open(device_name, O_RDWR | O_NONBLOCK);
         if (sg_fd < 0) {
+                err = errno;
                 perror("sg_test_rwbuf: open error");
-                return SG_LIB_FILE_ERROR;
+                return sg_convert_errno(err);
         }
         ret = find_out_about_buffer(sg_fd);
         if (ret)
@@ -489,7 +547,7 @@ int main (int argc, char * argv[])
                 goto err_out;
         }
 
-        cmpbuf = (unsigned char *)malloc(size);
+        cmpbuf = (uint8_t *)sg_memalign(size, 0, &free_cmpbuf, false);
         for (k = 0; k < times; ++k) {
                 ret = write_buffer (sg_fd, size);
                 if (ret) {
@@ -504,13 +562,13 @@ int main (int argc, char * argv[])
         }
 
 err_out:
-        if (cmpbuf)
-                free(cmpbuf);
+        if (free_cmpbuf)
+                free(free_cmpbuf);
         res = close(sg_fd);
         if (res < 0) {
                 perror(ME "close error");
                 if (0 == ret)
-                        ret = SG_LIB_FILE_ERROR;
+                        ret = sg_convert_errno(errno);
         }
         if ((0 == ret) && (! do_quick))
                 printf ("Success\n");
