@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2015 Douglas Gilbert
+ * Copyright (c) 2014-2018 Douglas Gilbert
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
@@ -17,6 +17,8 @@
 #include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
+#include <stdbool.h>
 #include <string.h>
 #include <errno.h>
 #include <limits.h>
@@ -29,13 +31,14 @@
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
+
 #include "sg_lib.h"
 #include "sg_pt.h"
 #include "sg_cmds_basic.h"
 #include "sg_unaligned.h"
 #include "sg_pr2serr.h"
 
-static const char * version_str = "1.08 20151220";
+static const char * version_str = "1.15 20180628";
 
 
 #define ME "sg_write_verify: "
@@ -118,24 +121,24 @@ usage()
 /* Invokes a SCSI WRITE AND VERIFY according with CDB. Returns 0 -> success,
  * various SG_LIB_CAT_* positive values or -1 -> other errors */
 static int
-run_scsi_transaction(int sg_fd, const unsigned char *cdbp, int cdb_len,
-                     unsigned char *dop, int do_len, int timeout, int verbose)
+run_scsi_transaction(int sg_fd, const uint8_t *cdbp, int cdb_len,
+                     uint8_t *dop, int do_len, int timeout,
+                     bool noisy, int verbose)
 {
     int res, k, sense_cat, ret;
-    unsigned char sense_b[SENSE_BUFF_LEN];
-    int noisy = 1;
     struct sg_pt_base * ptvp;
+    uint8_t sense_b[SENSE_BUFF_LEN];
     char b[32];
 
     snprintf(b, sizeof(b), "Write and verify(%d)", cdb_len);
     if (verbose) {
-       pr2serr("    %s cmd: ", b);
+       pr2serr("    %s cdb: ", b);
        for (k = 0; k < cdb_len; ++k)
            pr2serr("%02x ", cdbp[k]);
        pr2serr("\n");
        if ((verbose > 2) && dop && do_len) {
             pr2serr("    Data out buffer [%d bytes]:\n", do_len);
-            dStrHexErr((const char *)dop, do_len, -1);
+            hex2stderr(dop, do_len, -1);
         }
     }
     ptvp = construct_scsi_pt_obj();
@@ -147,10 +150,10 @@ run_scsi_transaction(int sg_fd, const unsigned char *cdbp, int cdb_len,
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
     set_scsi_pt_data_out(ptvp, dop, do_len);
     res = do_scsi_pt(ptvp, sg_fd, timeout, verbose);
-    ret = sg_cmds_process_resp(ptvp, b, res, 0, sense_b, noisy, verbose,
-                               &sense_cat);
+    ret = sg_cmds_process_resp(ptvp, b, res, SG_NO_DATA_IN, sense_b, noisy,
+                               verbose, &sense_cat);
     if (-1 == ret)
-        ;
+        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
     else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
@@ -159,7 +162,8 @@ run_scsi_transaction(int sg_fd, const unsigned char *cdbp, int cdb_len,
             break;
         case SG_LIB_CAT_MEDIUM_HARD:    /* write or verify failed */
             {
-                int valid, slen;
+                bool valid;
+                int slen;
                 uint64_t ull = 0;
 
                 slen = get_scsi_pt_sense_len(ptvp);
@@ -186,12 +190,13 @@ run_scsi_transaction(int sg_fd, const unsigned char *cdbp, int cdb_len,
 /* Invokes a SCSI WRITE AND VERIFY (10) command (SBC). Returns 0 -> success,
 * various SG_LIB_CAT_* positive values or -1 -> other errors */
 static int
-sg_ll_write_verify10(int sg_fd, int wrprotect, int dpo, int bytchk,
+sg_ll_write_verify10(int sg_fd, int wrprotect, bool dpo, int bytchk,
                      unsigned int lba, int num_lb, int group,
-                     unsigned char *dop, int do_len, int timeout, int verbose)
+                     uint8_t *dop, int do_len, int timeout,
+                     bool noisy, int verbose)
 {
     int ret;
-    unsigned char wv_cdb[WRITE_VERIFY10_CMDLEN];
+    uint8_t wv_cdb[WRITE_VERIFY10_CMDLEN];
 
     memset(wv_cdb, 0, WRITE_VERIFY10_CMDLEN);
     wv_cdb[0] = WRITE_VERIFY10_CMD;
@@ -205,19 +210,19 @@ sg_ll_write_verify10(int sg_fd, int wrprotect, int dpo, int bytchk,
     wv_cdb[6] = group & 0x1f;
     sg_put_unaligned_be16((uint16_t)num_lb, wv_cdb + 7);
     ret = run_scsi_transaction(sg_fd, wv_cdb, sizeof(wv_cdb), dop, do_len,
-                               timeout, verbose);
+                               timeout, noisy, verbose);
     return ret;
 }
 
 /* Invokes a SCSI WRITE AND VERIFY (16) command (SBC). Returns 0 -> success,
 * various SG_LIB_CAT_* positive values or -1 -> other errors */
 static int
-sg_ll_write_verify16(int sg_fd, int wrprotect, int dpo, int bytchk,
-                     uint64_t llba, int num_lb, int group, unsigned char *dop,
-                     int do_len, int timeout, int verbose)
+sg_ll_write_verify16(int sg_fd, int wrprotect, bool dpo, int bytchk,
+                     uint64_t llba, int num_lb, int group, uint8_t *dop,
+                     int do_len, int timeout, bool noisy, int verbose)
 {
     int ret;
-    unsigned char wv_cdb[WRITE_VERIFY16_CMDLEN];
+    uint8_t wv_cdb[WRITE_VERIFY16_CMDLEN];
 
 
     memset(wv_cdb, 0, sizeof(wv_cdb));
@@ -232,22 +237,25 @@ sg_ll_write_verify16(int sg_fd, int wrprotect, int dpo, int bytchk,
     sg_put_unaligned_be32((uint32_t)num_lb, wv_cdb + 10);
     wv_cdb[14] = group & 0x1f;
     ret = run_scsi_transaction(sg_fd, wv_cdb, sizeof(wv_cdb), dop, do_len,
-                               timeout, verbose);
+                               timeout, noisy, verbose);
     return ret;
 }
 
+/* Returns file descriptor ( >= 0) if successfule. Else a negated sg3_utils
+ * error code is returned. */
 static int
 open_if(const char * fn, int got_stdin)
 {
-    int fd;
+    int fd, err;
 
     if (got_stdin)
         fd = STDIN_FILENO;
     else {
         fd = open(fn, O_RDONLY);
         if (fd < 0) {
-            pr2serr(ME "open error: %s: %s\n", fn, safe_strerror(errno));
-            return -SG_LIB_FILE_ERROR;
+            err = errno;
+            pr2serr(ME "open error: %s: %s\n", fn, safe_strerror(err));
+            return -sg_convert_errno(err);
         }
     }
     if (sg_set_binary_mode(fd) < 0) {
@@ -260,31 +268,35 @@ open_if(const char * fn, int got_stdin)
 int
 main(int argc, char * argv[])
 {
-    int sg_fd, res, c, n, first_time;
-    unsigned char * wvb = NULL;
-    void * wrkBuff = NULL;
-    int dpo = 0;
+    bool do_16 = false;
+    bool dpo = false;
+    bool first_time;
+    bool given_do_16 = false;
+    bool has_filename = false;
+    bool lba_given = false;
+    bool repeat = false;
+    bool verbose_given = false;
+    bool version_given = false;
+    int sg_fd, res, c, n;
     int bytchk = 0;
     int group = 0;
-    int do_16 = 0;
-    int given_do_16 = 0;
-    uint64_t llba = 0;
-    int lba_given = 0;
-    uint32_t num_lb = 1;
-    uint32_t snum_lb = 1;
-    int repeat = 0;
-    int timeout = DEF_TIMEOUT_SECS;
-    int verbose = 0;
-    int64_t ll;
-    int wrprotect = 0;
-    const char * device_name = NULL;
-    const char * ifnp;
-    int has_filename = 0;
     int ilen = -1;
     int ifd = -1;
-    int ret = 1;
     int b_p_lb = 512;
+    int ret = 1;
+    int timeout = DEF_TIMEOUT_SECS;
     int tnum_lb_wr = 0;
+    int verbose = 0;
+    int wrprotect = 0;
+    uint32_t num_lb = 1;
+    uint32_t snum_lb = 1;
+    uint64_t llba = 0;
+    int64_t ll;
+    uint8_t * wvb = NULL;
+    uint8_t * wrkBuff = NULL;
+    uint8_t * free_wrkBuff = NULL;
+    const char * device_name = NULL;
+    const char * ifnp;
     char cmd_name[32];
 
     ifnp = "";          /* keep MinGW quiet */
@@ -307,12 +319,12 @@ main(int argc, char * argv[])
             }
             break;
         case 'd':
-            dpo = 1;
+            dpo = true;
             break;
         case 'g':
             group = sg_get_num(optarg);
-            if ((group < 0) || (group > 31))  {
-                pr2serr("argument to '--group' expected to be 0 to 31\n");
+            if ((group < 0) || (group > 63))  {
+                pr2serr("argument to '--group' expected to be 0 to 63\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
             break;
@@ -322,7 +334,7 @@ main(int argc, char * argv[])
             return 0;
         case 'i':
             ifnp = optarg;
-            has_filename = 1;
+            has_filename = true;
             break;
         case 'I':
             ilen = sg_get_num(optarg);
@@ -342,7 +354,7 @@ main(int argc, char * argv[])
                 return SG_LIB_SYNTAX_ERROR;
             }
             llba = (uint64_t)ll;
-            ++lba_given;
+            lba_given = true;
             break;
         case 'n':
             n = sg_get_num(optarg);
@@ -353,11 +365,11 @@ main(int argc, char * argv[])
             num_lb = (uint32_t)n;
             break;
         case 'R':
-            ++repeat;
+            repeat = true;
             break;
         case 'S':
-            do_16 = 1;
-            given_do_16 = 1;
+            do_16 = true;
+            given_do_16 = true;
             break;
         case 't':
             timeout = sg_get_num(optarg);
@@ -367,11 +379,12 @@ main(int argc, char * argv[])
             }
             break;
         case 'v':
+            verbose_given = true;
             ++verbose;
             break;
         case 'V':
-            pr2serr(ME "version: %s\n", version_str);
-            return 0;
+            version_given = true;
+            break;
         case 'w':
             wrprotect = sg_get_num(optarg);
             if ((wrprotect < 0) || (wrprotect > 7))  {
@@ -400,8 +413,29 @@ main(int argc, char * argv[])
        }
     }
 
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (verbose_given && version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        verbose_given = false;
+        version_given = false;
+        verbose = 0;
+    } else if (! verbose_given) {
+        pr2serr("set '-vv'\n");
+        verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", verbose);
+#else
+    if (verbose_given && version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (version_given) {
+        pr2serr(ME "version: %s\n", version_str);
+        return 0;
+    }
+
     if (NULL == device_name) {
-        pr2serr("missing device name!\n");
+        pr2serr("Missing device name!\n\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
@@ -414,12 +448,12 @@ main(int argc, char * argv[])
         if (! has_filename) {
             pr2serr("with '--repeat' need '--in=IF' option\n");
             usage();
-            return SG_LIB_SYNTAX_ERROR;
+            return SG_LIB_CONTRADICT;
         }
         if (ilen < 1) {
             pr2serr("with '--repeat' need '--ilen=ILEN' option\n");
             usage();
-            return SG_LIB_SYNTAX_ERROR;
+            return SG_LIB_CONTRADICT;
         } else {
             b_p_lb = ilen / num_lb;
             if (b_p_lb < 64) {
@@ -431,19 +465,20 @@ main(int argc, char * argv[])
         }
     }
 
-    sg_fd = sg_cmds_open_device(device_name, 0 /* rw */, verbose);
+    sg_fd = sg_cmds_open_device(device_name, false /* rw */, verbose);
     if (sg_fd < 0) {
+        ret = sg_convert_errno(-sg_fd);
         pr2serr(ME "open error: %s: %s\n", device_name, safe_strerror(-sg_fd));
-        return SG_LIB_FILE_ERROR;
+        goto err_out;
     }
 
-    if ((0 == do_16) && (llba > UINT_MAX))
-        do_16 = 1;
-    if ((0 == do_16) && (num_lb > 0xffff))
-        do_16 = 1;
+    if ((! do_16) && (llba > UINT_MAX))
+        do_16 = true;
+    if ((! do_16) && (num_lb > 0xffff))
+        do_16 = true;
     snprintf(cmd_name, sizeof(cmd_name), "Write and verify(%d)",
              (do_16 ? 16 : 10));
-    if (verbose && (0 == given_do_16) && do_16)
+    if (verbose && (! given_do_16) && do_16)
         pr2serr("Switching to %s because LBA or NUM too large\n", cmd_name);
     if (verbose) {
         pr2serr("Issue %s to device %s\n\tilen=%d", cmd_name, device_name,
@@ -452,10 +487,10 @@ main(int argc, char * argv[])
             pr2serr(" [0x%x]", ilen);
         pr2serr(", lba=%" PRIu64 " [0x%" PRIx64 "]\n\twrprotect=%d, dpo=%d, "
                 "bytchk=%d, group=%d, repeat=%d\n", llba, llba, wrprotect,
-                dpo, bytchk, group, repeat);
+                (int)dpo, bytchk, group, (int)repeat);
     }
 
-    first_time = 1;
+    first_time = true;
     do {
         if (first_time) {
             //If a file with data to write has been provided
@@ -491,12 +526,13 @@ main(int argc, char * argv[])
                     } else if (verbose)
                         pr2serr("Using file size of %d bytes\n", ilen);
                 }
-                if (NULL == (wrkBuff = malloc(ilen))) {
+                if (NULL == (wrkBuff = (uint8_t *)sg_memalign(ilen, 0,
+                                           &free_wrkBuff, verbose > 3))) {
                     pr2serr(ME "out of memory\n");
-                    ret = SG_LIB_CAT_OTHER;
+                    ret = sg_convert_errno(ENOMEM);
                     goto err_out;
                 }
-                wvb = (unsigned char *)wrkBuff;
+                wvb = (uint8_t *)wrkBuff;
                 res = read(ifd, wvb, ilen);
                 if (res < 0) {
                     pr2serr("Could not read from %s", ifnp);
@@ -506,8 +542,8 @@ main(int argc, char * argv[])
                     pr2serr("Read only %d bytes (expected %d) from %s\n", res,
                             ilen, ifnp);
                     if (repeat)
-                        pr2serr("Will scale subsequent pieces when repeat=1, "
-                                "but this is first\n");
+                        pr2serr("Will scale subsequent pieces when "
+                                "repeat=true, but this is first\n");
                     goto err_out;
                 }
             } else {
@@ -517,18 +553,19 @@ main(int argc, char * argv[])
                                 num_lb, 512, 512 * num_lb);
                     ilen = 512 * num_lb;
                 }
-                if (NULL == (wrkBuff = malloc(ilen))) {
+                if (NULL == (wrkBuff = (uint8_t *)sg_memalign(ilen, 0,
+                                           &free_wrkBuff, verbose > 3))) {
                     pr2serr(ME "out of memory\n");
-                    ret = SG_LIB_CAT_OTHER;
+                    ret = sg_convert_errno(ENOMEM);
                     goto err_out;
                 }
-                wvb = (unsigned char *)wrkBuff;
+                wvb = (uint8_t *)wrkBuff;
                 /* Not sure about this: default contents to 0xff bytes */
                 memset(wrkBuff, 0xff, ilen);
             }
-            first_time = 0;
+            first_time = false;
             snum_lb = num_lb;
-        } else {        /* repeat=1, first_time=0, must be reading file */
+        } else {    /* repeat=true, first_time=false, must be reading file */
             llba += snum_lb;
             res = read(ifd, wvb, ilen);
             if (res < 0) {
@@ -553,11 +590,12 @@ main(int argc, char * argv[])
         if (do_16)
             res = sg_ll_write_verify16(sg_fd, wrprotect, dpo, bytchk, llba,
                                        snum_lb, group, wvb, ilen, timeout,
-                                       verbose);
+                                       verbose > 0, verbose);
         else
             res = sg_ll_write_verify10(sg_fd, wrprotect, dpo, bytchk,
                                        (unsigned int)llba, snum_lb, group,
-                                       wvb, ilen, timeout, verbose);
+                                       wvb, ilen, timeout, verbose > 0,
+                                       verbose);
         ret = res;
         if (repeat && (0 == ret))
             tnum_lb_wr += snum_lb;
@@ -569,23 +607,20 @@ err_out:
     if (repeat)
         pr2serr("%d [0x%x] logical blocks written, in total\n", tnum_lb_wr,
                 tnum_lb_wr);
-    if (wrkBuff)
-        free(wrkBuff);
+    if (free_wrkBuff)
+        free(free_wrkBuff);
     if ((ifd >= 0) && (STDIN_FILENO != ifd))
         close(ifd);
     res = sg_cmds_close_device(sg_fd);
     if (res < 0) {
         pr2serr("close error: %s\n", safe_strerror(-res));
         if (0 == ret)
-            return SG_LIB_FILE_ERROR;
+            ret = sg_convert_errno(-res);
     }
     if (ret && (0 == verbose)) {
-        if (SG_LIB_CAT_INVALID_OP == ret)
-            pr2serr("%s command not supported\n", cmd_name);
-        else if (ret > 0)
-            pr2serr("%s, exit status %d\n", cmd_name, ret);
-        else if (ret < 0)
-            pr2serr("Some error occurred\n");
+        if (! sg_if_can2stderr("sg_write_verify failed: ", ret))
+            pr2serr("Some error occurred, try again with '-v' "
+                    "or '-vv' for more information\n");
     }
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
